import pytest

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.weave_client import get_ref

from ...trace_server import trace_server_interface as tsi


def test_op_return_sync_empty(client):
    @weave.op()
    def fn():
        return

    fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output is None


@pytest.mark.asyncio
async def test_op_return_async_empty(client):
    @weave.op()
    async def fn():
        return

    await fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output is None


def test_op_return_sync_obj(client):
    @weave.op()
    def fn():
        return 1

    fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == 1


@pytest.mark.asyncio
async def test_op_return_async_obj(client):
    @weave.op()
    async def fn():
        return 1

    await fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == 1


def simple_list_accumulator(acc, value):
    if acc is None:
        acc = []
    acc.append(value)
    return acc


""" There are 16 tests that follow, permutations of the following:
    - (2x) sync/async
    - (2x) generator/iterator
    - (4x) never iterated/partially iterated/fully iterated/exception thrown
"""


def test_op_return_sync_generator(client):
    @weave.op()
    def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    for item in fn():
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, -1, -1))


@pytest.mark.asyncio
async def test_op_return_async_generator(client):
    @weave.op()
    async def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    async for item in fn():
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, -1, -1))


def test_op_return_sync_iterator(client):
    class MyIterator:
        size = 10

        def __iter__(self):
            return self

        def __next__(self):
            if self.size == 0:
                raise StopIteration
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    for item in fn():
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, -1, -1))


@pytest.mark.asyncio
async def test_op_return_async_iterator(client):
    class MyAsyncIterator:
        size = 10

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.size == 0:
                raise StopAsyncIteration
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyAsyncIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    async for item in fn():
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, -1, -1))


def test_op_return_sync_generator_never_iter(client):
    @weave.op()
    def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output is None


@pytest.mark.asyncio
async def test_op_return_async_generator_never_iter(client):
    @weave.op()
    async def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output is None


def test_op_return_sync_iterator_never_iter(client):
    class MyIterator:
        size = 10

        def __iter__(self):
            return self

        def __next__(self):
            if self.size == 0:
                raise StopIteration
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output is None


@pytest.mark.asyncio
async def test_op_return_async_iterator_never_iter(client):
    class MyAsyncIterator:
        size = 10

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.size == 0:
                raise StopAsyncIteration
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyAsyncIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    fn()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output is None


def test_op_return_sync_generator_partial(client):
    @weave.op()
    def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    for item in fn():
        if item == 5:
            break

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))


@pytest.mark.asyncio
async def test_op_return_async_generator_partial(client):
    @weave.op()
    async def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    async for item in fn():
        if item == 5:
            break

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))


def test_op_return_sync_iterator_partial(client):
    class MyIterator:
        size = 10

        def __iter__(self):
            return self

        def __next__(self):
            if self.size == 0:
                raise StopIteration
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    for item in fn():
        if item == 5:
            break

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))


@pytest.mark.asyncio
async def test_op_return_async_iterator_partial(client):
    class MyAsyncIterator:
        size = 10

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.size == 0:
                raise StopAsyncIteration
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyAsyncIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    async for item in fn():
        if item == 5:
            break

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))


def test_op_return_sync_generator_exception(client):
    @weave.op()
    def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size
            if size == 5:
                raise Exception("test")

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    try:
        for item in fn():
            pass
    except Exception:
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))
    assert res.calls[0].exception != None


@pytest.mark.asyncio
async def test_op_return_async_generator_exception(client):
    @weave.op()
    async def fn():
        size = 10
        while size > 0:
            size -= 1
            yield size
            if size == 5:
                raise Exception("test")

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    try:
        async for item in fn():
            pass
    except Exception:
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))
    assert res.calls[0].exception != None


def test_op_return_sync_iterator_exception(client):
    class MyIterator:
        size = 10

        def __iter__(self):
            return self

        def __next__(self):
            if self.size == 0:
                raise StopIteration
            if self.size == 5:
                raise Exception("test")
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    try:
        for item in fn():
            pass
    except Exception:
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))
    assert res.calls[0].exception != None


@pytest.mark.asyncio
async def test_op_return_async_iterator_exception(client):
    class MyAsyncIterator:
        size = 10

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.size == 0:
                raise StopAsyncIteration
            if self.size == 5:
                raise Exception("test")
            self.size -= 1
            return self.size

    @weave.op()
    def fn():
        return MyAsyncIterator()

    add_accumulator(fn, lambda inputs: simple_list_accumulator)

    try:
        async for item in fn():
            pass
    except Exception:
        pass

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    obj_ref = get_ref(fn)
    assert obj_ref is not None
    assert res.calls[0].op_name == obj_ref.uri()
    assert res.calls[0].inputs == {}
    assert res.calls[0].output == list(range(9, 4, -1))
    assert res.calls[0].exception != None



import typing

from weave.legacy.weave import client_interface, context_state, errors, graph
from weave.legacy.weave import weave_types as types


def dereference_variables(
    node: graph.Node, var_values: graph.Frame, missing_ok: bool = False
) -> graph.Node:
    def map_fn(n: graph.Node) -> graph.Node:
        if isinstance(n, graph.VarNode):
            try:
                return var_values[n.name]
            except KeyError:
                pass
                # if not missing_ok:
                #     raise errors.WeaveMissingVariableError(n.name)
        return n

    return graph.map_nodes_top_level([node], map_fn)[0]


def call_fn(
    weave_fn: graph.Node,
    inputs: graph.Frame,
) -> graph.Node:
    return dereference_variables(weave_fn, inputs)


def better_call_fn(weave_fn: graph.ConstNode, *inputs: graph.Node) -> graph.Node:
    call_inputs = {}
    if not isinstance(weave_fn.type, types.Function):
        raise errors.WeaveInternalError(
            "Expected function type, got %s" % weave_fn.type
        )
    for input_name, input in zip(weave_fn.type.input_types.keys(), inputs):
        call_inputs[input_name] = input
    res = call_fn(weave_fn.val, call_inputs)  # type: ignore
    # Convert to Runtime nodes so dispatch works.
    if isinstance(res, graph.OutputNode):
        return make_output_node(res.type, res.from_op.name, res.from_op.inputs)  # type: ignore
    elif isinstance(res, graph.VarNode):
        return make_var_node(res.type, res.name)
    elif isinstance(res, graph.ConstNode):
        return make_const_node(res.type, res.val)
    else:
        raise errors.WeaveInternalError("Invalid Node: %s" % res)


def use(
    nodes: typing.Union[graph.Node, typing.Sequence[graph.Node]],
    client: typing.Union[client_interface.ClientInterface, None] = None,
) -> typing.Union[typing.Any, typing.Sequence[typing.Any]]:
    if client is None:
        client = context_state.get_client()
        if client is None:
            raise errors.WeaveInternalError("no client set")
    single = True
    if not isinstance(nodes, graph.Node):
        single = False
    else:
        nodes = [nodes]

    # Do this to convert all Refs to get(str(ref))
    # But this is incorrect! If there are shared parent nodes among nodes, we will
    # disconnect them.
    # TODO: Fix
    actual_nodes = []
    for node in nodes:
        if not isinstance(node, graph.Node):
            if context_state.eager_mode():
                raise errors.WeaveApiError("use not allowed in eager mode.")
            else:
                raise errors.WeaveApiError("non-Node passed to use(): %s" % type(node))
        actual_nodes.append(node)

    result = client.execute(actual_nodes)

    if single:
        result = result[0]
    return result


def make_var_node(type_: types.Type, name: str) -> graph.VarNode:
    # Circular import. TODO: fix
    from weave.legacy.weave import dispatch

    return dispatch.RuntimeVarNode(type_, name)


def make_const_node(type_: types.Type, val: typing.Any) -> graph.ConstNode:
    # Circular import. TODO: fix
    from weave.legacy.weave import dispatch

    return dispatch.RuntimeConstNode(type_, val)


def const(val: typing.Any, type: typing.Optional[types.Type] = None) -> graph.ConstNode:
    if type is None:
        type = types.TypeRegistry.type_of(val)
    return make_const_node(type, val)


def make_var_for_value(v: typing.Any, name: str) -> graph.VarNode:
    """Make a VarNode whose value is v."""
    if not isinstance(v, graph.Node):
        v = const(v)
    new_var = make_var_node(v.type, name)
    new_var._var_val = v
    return new_var


def make_output_node(
    type_: types.Type, op_name: str, op_params: dict[str, graph.Node]
) -> graph.OutputNode:
    # Circular import. TODO: fix
    from weave.legacy.weave import dispatch

    return dispatch.RuntimeOutputNode(type_, op_name, op_params)


# Given a registered op, make a mapped version of it.
def define_fn(
    parameters: dict[str, types.Type], body: typing.Callable[..., graph.Node]
) -> graph.ConstNode:
    var_nodes = [make_var_node(t, k) for k, t in parameters.items()]
    try:
        from weave.legacy.weave import op_def

        with op_def.no_refine():
            fnNode = body(*var_nodes)
    except errors.WeaveExpectedConstError as e:
        raise errors.WeaveMakeFunctionError("function body expected const node.")
    if not isinstance(fnNode, graph.Node):
        raise errors.WeaveMakeFunctionError("output_type function must return a node.")
    return const(fnNode, types.Function(parameters, fnNode.type))


ENVType = typing.Union[graph.Node, typing.Callable[..., graph.Node]]
ENInputTypeType = typing.Optional[
    typing.Union[types.Type, typing.Callable[..., types.Type]]
]
ENBoundParamsType = typing.Optional[dict[str, graph.Node]]


# this refines all the nodes in a graph. it's useful / needed if you do call_fn on some arguments
# that have a different type than exact type of the function inputs, e.g., the argument is tagged
# and the function doesn't explicitly operate on tagged values. this ensures that the input tags
# are propagated appropriately to the output type of the function.
def refine_graph(node: graph.Node) -> graph.Node:
    from weave.legacy.weave.registry_mem import memory_registry

    if isinstance(node, (graph.ConstNode, graph.VoidNode, graph.VarNode)):
        return node
    elif isinstance(node, graph.OutputNode):
        op_name = node.from_op.name
        op = memory_registry.get_op(op_name)
        if op is None:
            raise ValueError("No op named %s" % op_name)
        refined_inputs: dict[str, typing.Any] = {}
        for input_name, input_node in node.from_op.inputs.items():
            refined_input = refine_graph(input_node)
            refined_inputs[input_name] = refined_input
        return op(**refined_inputs)

    else:
        raise NotImplementedError(
            "refine_graph cannot yet handle nodes of type %s" % type(node)
        )


def manual_call(
    op_name: str, inputs: dict[str, graph.Node], output_type: types.Type
) -> graph.Node:
    """Produce an output node manually.

    You can produce incorrect nodes this way. Use with caution.
    """
    from weave.legacy.weave import dispatch

    return dispatch.RuntimeOutputNode(output_type, op_name, inputs)



import typing
from itertools import chain

import numpy as np
import pyarrow as pa
import pytest

import weave
from weave.legacy.weave import api, ops, storage, weave_internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.ops_arrow import ArrowWeaveList, arrow_as_array
from weave.legacy.weave.ops_domain.run_segment import RunSegment

N_NUMERIC_METRICS = 99  # number of numerical columns in the metrics table


def random_metrics(n: int = 10, starting_step: int = 0, delta_step: int = 1):
    """Create an array of metrics of length n starting from step starting_index."""
    if n <= 0:
        raise ValueError("n must be at least 1")
    if starting_step < 0:
        raise ValueError("starting index must be at least 0")
    if delta_step < 1:
        raise ValueError("delta_step must be an integer greater than or equal to 1.")

    data = {
        "step": np.arange(starting_step, starting_step + n * delta_step, delta_step),
        "string_col": np.random.choice(list("ABCDEFGHIJKLMNOPQRSTUVWXYZ"), n),
    }
    for j in range(N_NUMERIC_METRICS):
        data[f"metric{j}"] = np.random.random(n)
    return ArrowWeaveList(pa.table(data))


def create_branch(
    name: str,
    previous_segment: typing.Optional[RunSegment] = None,
    length=10,
    previous_segment_branch_frac=0.8,
) -> RunSegment:
    """Create a new segment and optionally attach it to a previous segment.

    Parameters
    ----------
    name: str
       The name of the segment.
    previous_segment: Optional[RunSegment], default None.
       The parent run segment. If this is a root run segment, use None.
    length: int, default = 10
       The number of history rows to generate for the segment.
    previous_segment_branch_frac: float satisfying 0 < branch_frac <= 1.
       Parameter describing where in the previous segment to set the branch point.
       A previous_segment_branch_frac of 0 sets the branch point at the previous
       segment's root, whereas a previous_segment_branch_frac of 1 sets the branch
       point at the end of the previous segment. A previous_segment_branch_frac of
       0.5 would include half of the previous segment's metric rows.

    Returns
    -------
    segment: RunSegment
        The new segment.
    """
    if not (0 < previous_segment_branch_frac <= 1):
        raise ValueError("branch_frac must satisfy 0 < branch_frac <= 1")

    if length <= 0:
        raise ValueError("Length must be greater than 0.")

    if previous_segment:
        previous_metrics = previous_segment.metrics
        n_previous_metrics = len(previous_metrics)
        if n_previous_metrics > 0:
            previous_segment_branch_index = (
                int(previous_segment_branch_frac * n_previous_metrics) - 1
            )

            # this run segment has a different root than the previous one
            if previous_segment_branch_index < 0:
                raise ValueError(
                    f"Invalid branch point on RunSegment: previous_segment_branch_index "
                    f"{previous_segment_branch_index} must be between 0 and {len(previous_metrics) - 1}"
                )

            previous_segment_branch_step = (
                previous_metrics._index(0)["step"] + previous_segment_branch_index
            )

            ref = storage.save(previous_segment)
            new_metrics = random_metrics(
                n=length, starting_step=previous_segment_branch_step + 1
            )

            return RunSegment(name, new_metrics, ref.uri, previous_segment_branch_index)
    return RunSegment(name, random_metrics(length, 0), None, 0)


def create_experiment(
    num_steps: int, num_runs: int, branch_frac: float = 0.8
) -> typing.Optional[RunSegment]:
    num_steps_per_run = num_steps // num_runs
    segment = None
    for i in range(num_runs):
        segment = create_branch(
            f"branch {i}",
            segment,
            length=num_steps_per_run,
            previous_segment_branch_frac=branch_frac,
        )
    return segment


@pytest.fixture()
def num_steps():
    return 100


@pytest.fixture()
def num_runs():
    return 20


def get_awl_col(awl: ArrowWeaveList, col_name: str):
    arr = arrow_as_array(awl._arrow_data)
    return arr.field(col_name)


@pytest.mark.parametrize("branch_frac", [0.0, 0.8, 1.0])
def test_experiment_branching(branch_frac, num_steps, num_runs):
    steps_per_run = num_steps // num_runs

    try:
        segment = create_experiment(num_steps, num_runs, branch_frac)
    except ValueError:
        assert branch_frac == 0
    else:
        storage.save(segment)
        experiment = api.use(segment.experiment())
        assert (
            len(experiment)
            == int(steps_per_run * branch_frac) * (num_runs - 1) + steps_per_run
        )

        assert (
            get_awl_col(experiment, "step").to_pylist()
            == list(range(int(steps_per_run * branch_frac) * (num_runs - 1)))
            + get_awl_col(segment.metrics, "step").to_pylist()
        )


@pytest.mark.parametrize("delta_step", [1, 2, 3])
def test_explicit_experiment_construction(delta_step):
    root_segment = RunSegment(
        "my-first-run", random_metrics(10, delta_step=delta_step), None, 0
    )
    ref1 = storage.save(root_segment)
    segment1 = RunSegment(
        "my-second-run",
        random_metrics(10, 5 * delta_step, delta_step=delta_step),
        ref1.uri,
        4,
    )
    ref2 = storage.save(segment1)
    segment2 = RunSegment(
        "my-third-run",
        random_metrics(5, 10 * delta_step, delta_step=delta_step),
        ref2.uri,
        4,
    )
    storage.save(segment2)
    experiment = api.use(segment2.experiment())

    assert get_awl_col(experiment, "step").to_pylist() == list(
        range(0, 15 * delta_step, delta_step)
    )
    assert (
        get_awl_col(experiment, "string_col").to_pylist()
        == get_awl_col(root_segment.metrics, "string_col").to_pylist()[:5]
        + get_awl_col(segment1.metrics, "string_col").to_pylist()[:5]
        + get_awl_col(segment2.metrics, "string_col").to_pylist()
    )

    assert get_awl_col(experiment, "run_name").to_pylist() == list(
        chain(
            *[[name] * 5 for name in ["my-first-run", "my-second-run", "my-third-run"]]
        )
    )


def test_invalid_explicit_experiment_construction():
    root_segment = RunSegment("my-first-run", random_metrics(10))
    ref1 = storage.save(root_segment)

    # this run has no metrics
    segment1 = RunSegment(
        "my-second-run",
        root_segment.metrics._limit(0),
        ref1.uri,
        4,
    )
    ref2 = storage.save(segment1)

    # this run tries to branch off a run with no metrics, which is not possible
    segment2 = RunSegment(
        "my-third-run",
        random_metrics(5, 10),
        ref2.uri,
        5,
    )
    storage.save(segment2)

    with pytest.raises(ValueError):
        api.use(segment2.experiment())


def test_vectorized_unnest_list_for_panelplot():
    metrics = random_metrics(10)
    root_segment = weave.save(RunSegment("my-first-run", metrics))

    def map_fn(row):
        return ops.dict_(
            **{
                "100": 100,
                "step": row["step"],
                "metric0": row["metric0"],
                "string_col": row["string_col"],
                "circle": "circle",
            }
        )

    fn_node = ops.define_fn(
        {
            "row": types.TypedDict(
                {
                    "100": types.Int(),
                    "step": types.Int(),
                    "metric0": types.Float(),
                    "string_col": types.String(),
                    "circle": types.String(),
                }
            )
        },
        map_fn,
    )

    res = weave.use(root_segment.metrics.map(fn_node))
    mapped = res.to_pylist_raw()
    metrics_arr = metrics._arrow_data.to_pylist()
    assert mapped == [
        {
            "100": 100,
            "circle": "circle",
            "string_col": metrics_arr[i]["string_col"],
            "metric0": metrics_arr[i]["metric0"],
            "step": i,
        }
        for i in range(len(metrics))
    ]


@pytest.fixture()
def number_bin_fn_node():
    return ops.numbers_bins_equal([1, 2, 3, 4], 10)


def test_number_bin_fn_node_type(number_bin_fn_node):
    assert number_bin_fn_node.type == types.Function(
        input_types={"row": types.Number()},
        output_type=types.NumberBinType,
    )


def test_number_bin_generation(number_bin_fn_node):
    # extract the function from its containing node
    function = api.use(number_bin_fn_node)
    call_node = ops.call_fn(
        function, {"row": weave_internal.make_const_node(types.Number(), 2.5)}
    )
    result = api.use(call_node)

    assert np.isclose(result["start"], 2.4)
    assert np.isclose(result["stop"], 2.7)


def test_number_bin_assignment_in_bin_range(number_bin_fn_node):
    # create a graph representing bin assignment
    assigned_number_bin_node = ops.number_bin(in_=2.5, bin_fn=number_bin_fn_node)
    assigned_bin = api.use(assigned_number_bin_node)

    assert np.isclose(assigned_bin["start"], 2.4)
    assert np.isclose(assigned_bin["stop"], 2.7)


def test_number_bin_assignment_outside_bin_range(number_bin_fn_node):
    # now do one outside the original range
    assigned_number_bin_node = ops.number_bin(in_=7, bin_fn=number_bin_fn_node)
    assigned_bin = api.use(assigned_number_bin_node)

    assert np.isclose(assigned_bin["start"], 6.9)
    assert np.isclose(assigned_bin["stop"], 7.2)


def test_group_by_bins_arrow_vectorized():
    number_bin_fn_node = ops.numbers_bins_equal([0, 10], 2)
    segment = create_experiment(200, 5, 0.8)

    def groupby_func(row):
        step = row["step"]
        assigned_number_bin_node = ops.number_bin(in_=step, bin_fn=number_bin_fn_node)
        return ops.dict_(number_bin_col_name=assigned_number_bin_node)

    func_node = weave_internal.define_fn(
        {"row": api.type_of(segment.metrics).object_type}, groupby_func
    )
    groupby_node = weave_internal.const(segment.metrics).groupby(func_node)

    result = api.use(groupby_node)
    assert api.use(weave_internal.const(result).count()) == 9

    group_key_node = weave_internal.const(result)[4].groupkey()
    key = api.use(group_key_node)
    assert key == {"number_bin_col_name": {"start": 145.0, "stop": 150.0}}


# cache busting didn't work properly with this query before
def test_map_merge_cache_busting():
    root_segment = RunSegment("my-first-run", random_metrics(10))
    ref = storage.save(root_segment)

    def map_fn_1_body(row):
        const_dict = ops.dict_()
        merge_dict = ops.dict_(
            **{
                "100": 100,
                "step": row["step"],
                "metric0": row["metric0"],
                "string_col": row["string_col"],
                "circle": "circle",
            }
        )
        return const_dict.merge(merge_dict)

    fn_node = weave_internal.define_fn(
        {"row": root_segment.metrics.object_type}, map_fn_1_body
    )
    query = api.get(ref).metrics.map(fn_node)
    result1 = api.use(query)

    def map_fn_2_body(row):
        const_dict = ops.dict_()
        merge_dict = ops.dict_(
            **{
                "100": 100,
                "step": row["step"],
                "metric1": row["metric1"],
                "string_col": row["string_col"],
                "circle": "circle",
            }
        )
        return const_dict.merge(merge_dict)

    fn_node = weave_internal.define_fn(
        {"row": root_segment.metrics.object_type}, map_fn_2_body
    )
    query = api.get(ref).metrics.map(fn_node)
    result2 = api.use(query)

    assert result1._arrow_data != result2._arrow_data
    assert (
        result1._arrow_data.type.get_field_index("metric0") != -1
        and result2._arrow_data.type.get_field_index("metric0") == -1
    )
    assert (
        result1._arrow_data.type.get_field_index("metric1") == -1
        and result2._arrow_data.type.get_field_index("metric1") != -1
    )


@pytest.mark.skip()  # TODO(dg): enable
def test_map_experiment_profile_post_groupby_map():
    last_segment = create_experiment(500000, 20)
    experiment = last_segment.experiment()

    group_key_name = "steppybin(pybinsequal (list (2, 500) , 2) )"
    list_node = ops.list_.make_list(**{"0": 2, "1": 500})
    number_bin_fn_node = ops.numbers_bins_equal(list_node, 2)

    def groupby_fn(row):
        step = row.pick("step")
        assigned_number_bin_node = ops.number_bin(in_=step, bin_fn=number_bin_fn_node)
        return ops.dict_(**{group_key_name: assigned_number_bin_node})

    groupby_node = weave_internal.define_fn(
        {"row": experiment.type.object_type}, groupby_fn
    )
    groupby = experiment.groupby(groupby_node)

    def map_fn_1_body(row):
        row_key = ops.WeaveGroupResultInterface.key(row)
        merge_dict = ops.dict_(
            **{
                "100": 100,
                "step": row.pick("step"),
                "metric0": row.pick("metric0"),
                "string_col": row.pick("string_col"),
                "circle": "circle",
            }
        )
        return row_key.merge(merge_dict)

    map_fn_node = weave_internal.define_fn(
        {
            "row": ops.arrow.awl_group_by_result_object_type(
                experiment.type.object_type, groupby_node.type.output_type
            )
        },
        map_fn_1_body,
    )
    mapped = ops.list_.make_list(**{"0": ops.list_.unnest(groupby.map(map_fn_node))})

    import cProfile

    cProfile.runctx("use(mapped)", globals(), locals(), filename="map_profile.pstat")



from typing import Any

from pydantic import field_validator

import weave
from weave.flow.obj import Object
from weave.trace.vals import WeaveTable


def short_str(obj: Any, limit: int = 25) -> str:
    str_val = str(obj)
    if len(str_val) > limit:
        return str_val[:limit] + "..."
    return str_val


class Dataset(Object):
    """
    Dataset object with easy saving and automatic versioning

    Examples:

    ```python
    # Create a dataset
    dataset = Dataset(name='grammar', rows=[
        {'id': '0', 'sentence': "He no likes ice cream.", 'correction': "He doesn't like ice cream."},
        {'id': '1', 'sentence': "She goed to the store.", 'correction': "She went to the store."},
        {'id': '2', 'sentence': "They plays video games all day.", 'correction': "They play video games all day."}
    ])

    # Publish the dataset
    weave.publish(dataset)

    # Retrieve the dataset
    dataset_ref = weave.ref('grammar').get()

    # Access a specific example
    example_label = dataset_ref.rows[2]['sentence']
    ```
    """

    rows: weave.Table

    @field_validator("rows", mode="before")
    def convert_to_table(cls, rows: Any) -> weave.Table:
        if not isinstance(rows, weave.Table):
            table_ref = getattr(rows, "table_ref", None)
            if isinstance(rows, WeaveTable):
                rows = list(rows)
            rows = weave.Table(rows)
            if table_ref:
                rows.table_ref = table_ref
        if len(rows.rows) == 0:
            raise ValueError("Attempted to construct a Dataset with an empty list.")
        for row in rows.rows:
            if not isinstance(row, dict):
                raise ValueError(
                    "Attempted to construct a Dataset with a non-dict object. Found type: "
                    + str(type(row))
                    + " of row: "
                    + short_str(row)
                )
            if len(row) == 0:
                raise ValueError(
                    "Attempted to construct a Dataset row with an empty dict."
                )
        return rows



import math
import typing

from weave.legacy.weave import api as weave
from weave.legacy.weave import context, context_state


class XOnly(typing.TypedDict):
    x: float


class Point(typing.TypedDict):
    x: float
    y: float


_loading_builtins_token = context_state.set_loading_built_ins()


@weave.op()
def _test_compute_points_compute_points(xs: list[XOnly], freq: float) -> list[Point]:
    res: list[Point] = []
    for row in xs:
        res.append({"x": row["x"], "y": math.sin(freq * row["x"])})
    return res


context_state.clear_loading_built_ins(_loading_builtins_token)


def test_compute_points():
    xs = [{"x": float(i)} for i in range(2)]
    points = _test_compute_points_compute_points(xs, 1)
    with context.local_http_client():
        assert weave.use(points) == [
            {"x": 0.0, "y": 0.0},
            {"x": 1.0, "y": 0.8414709848078965},
        ]



# Test script that downloads a bunch of files from a wandb artifact.
# Shows how to configure the various async APIs we use for WeaveIOManager.
# I had this up to 25MB/s download speed on the dsviz artifact (on both
# a gcp machine and a local network). Its probably slower now as I've
# added a bunch of features to the libraries without testing perf at
# this level, like safe file download and renaming.
#
# Try setting http.TRACE=True when you run this.
#
# Run from repo root with: `python -m weave.test_scripts.wandb_artifact_perf`

import asyncio
import cProfile
import time

from weave.legacy.weave import (
    artifact_wandb,
    async_map,
    engine_trace,
    filesystem,
    wandb_api,
    wandb_file_manager,
    weave_http,
)

tracer = engine_trace.tracer()  # type: ignore


async def gql_test() -> None:
    api = await wandb_api.get_wandb_api()
    fs = filesystem.get_filesystem_async()
    net = weave_http.HttpAsync(fs)
    file_man = wandb_file_manager.WandbFileManagerAsync(fs, net, api)
    man = await file_man.manifest(
        artifact_wandb.WeaveWBArtifactURI(
            "raw_data",
            "v4",
            entity_name="shawn",
            project_name="dsviz_demo",
        )
    )
    if man is None:
        raise Exception("Manifest is None")
    result_paths = []
    start_time = time.time()
    with cProfile.Profile() as profile:
        paths = list(man.get_paths_in_directory(""))[:1000]
        uris = [
            artifact_wandb.WeaveWBArtifactURI(
                "raw_data",
                "v4",
                entity_name="shawn",
                project_name="dsviz_demo",
                path=p,
            )
            for p in paths
        ]
        result_paths = await async_map.map_with_parallel_workers(
            uris, file_man.ensure_file, max_parallel=200
        )
    total_time = time.time() - start_time
    total_size = 0

    print("Downloaded N files:", len(result_paths))
    for path in result_paths:
        if path is None:
            raise Exception("Got none path")
        total_size += await fs.getsize(path)
    print(
        f"Downloaded {total_size} bytes in {total_time}s. {total_size / (1024**2) / total_time} MB/s"
    )

    profile.dump_stats("/tmp/profile.prof")


if __name__ == "__main__":
    asyncio.run(gql_test())



from typing import Any, Optional

from weave.trace.refs import ObjectRef, OpRef, parse_uri


class ObjectRefStrMatcher:
    def __init__(
        self,
        entity: Optional[str] = None,
        project: Optional[str] = None,
        kind: Optional[str] = None,
        name: Optional[str] = None,
        digest: Optional[str] = None,
        extra: Optional[list[str]] = None,
    ) -> None:
        self.entity = entity
        self.project = project
        self.kind = kind
        self.name = name
        self.digest = digest
        self.extra = extra

    def __eq__(self, other: Any) -> bool:
        other_ref = parse_uri(other)
        if not isinstance(other_ref, ObjectRef):
            return False
        if self.entity is not None:
            if self.entity != other_ref.entity:
                return False
        if self.project is not None:
            if self.project != other_ref.project:
                return False
        if self.kind is not None:
            if isinstance(other_ref, ObjectRef):
                other_kind = "object"
            elif isinstance(other_ref, OpRef):
                other_kind = "op"
            else:
                raise ValueError(f"Unknown kind: {other_ref}")
            if self.kind != other_kind:
                return False
        if self.name is not None:
            if self.name != other_ref.name:
                return False
        if self.digest is not None:
            if self.digest != other_ref.digest:
                return False
        if self.extra is not None:
            if self.extra != other_ref.extra:
                return False
        return True



# Official interface for making http requests. All weave
# interactions with http servers should go through this interface.

import asyncio
import logging
import os
import time
import types
import typing

import aiohttp
import requests
import requests.auth
import yarl

from weave.legacy.weave import engine_trace, filesystem, server_error_handling

logging.getLogger("aiohttp.access").setLevel(logging.WARNING)
logging.getLogger("aiohttp.client").setLevel(logging.WARNING)
logging.getLogger("aiohttp.internal").setLevel(logging.WARNING)
logging.getLogger("aiohttp.server").setLevel(logging.WARNING)
logging.getLogger("aiohttp.web").setLevel(logging.WARNING)
logging.getLogger("aiohttp.websocket").setLevel(logging.WARNING)


tracer = engine_trace.tracer()  # type: ignore

# Turn on to print stats for each request
ENABLE_REQUEST_TRACING = False


def logging_trace_config() -> aiohttp.TraceConfig:
    def make_trace_event_handler(
        event_name: str,
    ) -> typing.Callable[
        [aiohttp.ClientSession, types.SimpleNamespace, typing.Any],
        typing.Coroutine[None, None, None],
    ]:
        async def on_event(
            session: aiohttp.ClientSession,
            trace_config_ctx: types.SimpleNamespace,
            params: typing.Any,
        ) -> None:
            if not hasattr(trace_config_ctx, "events"):
                trace_config_ctx.events = []
            trace_config_ctx.events.append(
                (event_name, asyncio.get_event_loop().time())
            )

        return on_event

    async def on_request_end(
        session: aiohttp.ClientSession,
        trace_config_ctx: types.SimpleNamespace,
        params: aiohttp.tracing.TraceRequestEndParams,
    ) -> None:
        trace_config_ctx.events.append(("redirect", asyncio.get_event_loop().time()))
        trace_config_ctx.events.append(("req_end", asyncio.get_event_loop().time()))
        events = trace_config_ctx.events
        prior_e = events[0]
        s = f"{prior_e[0]}"
        for e in events[1:]:
            s += f" [{e[1]-prior_e[1]:.3f}s] {e[0]}"
            prior_e = e
        print("Req done: ", time.time(), s)

    trace_config = aiohttp.TraceConfig()
    trace_config.on_connection_queued_start.append(
        make_trace_event_handler("queued_start")
    )
    trace_config.on_connection_queued_end.append(make_trace_event_handler("queued_end"))
    trace_config.on_connection_create_start.append(
        make_trace_event_handler("conn_create_start")
    )
    trace_config.on_connection_create_end.append(
        make_trace_event_handler("conn_create_end")
    )
    trace_config.on_request_start.append(make_trace_event_handler("request_start"))
    trace_config.on_request_start.append(make_trace_event_handler("request_redirect"))
    trace_config.on_request_end.append(on_request_end)
    return trace_config


class HttpAsync:
    def __init__(self, fs: filesystem.FilesystemAsync) -> None:
        self.fs = fs

        conn = aiohttp.TCPConnector(limit=50)
        trace_configs = []
        if ENABLE_REQUEST_TRACING:
            trace_configs.append(logging_trace_config())
        self.session = aiohttp.ClientSession(
            trace_configs=trace_configs,
            connector=conn,
            cookie_jar=aiohttp.DummyCookieJar(),
        )

    async def __aenter__(self) -> "HttpAsync":
        return self

    async def __aexit__(self, *args: typing.Any) -> None:
        await self.session.close()

    async def download_file(
        self,
        url: str,
        path: str,
        headers: typing.Optional[dict[str, str]] = None,
        cookies: typing.Optional[dict[str, str]] = None,
        auth: typing.Optional[aiohttp.BasicAuth] = None,
    ) -> None:
        await self.fs.makedirs(os.path.dirname(path), exist_ok=True)
        with tracer.trace("download_file_task"):
            # TODO: Error handling when no file or manifest

            # yarl.URL encoded=True is very important! Otherwise aiohttp
            # will encode the url again and we'll get a 404 for things like
            # signed URLs
            async with self.session.get(
                yarl.URL(url, encoded=True), headers=headers, cookies=cookies, auth=auth
            ) as r:
                if r.status == 200:
                    async with self.fs.open_write(path, mode="wb") as f:
                        async for data in r.content.iter_chunked(16 * 1024):
                            await f.write(data)
                else:
                    raise server_error_handling.WeaveInternalHttpException.from_code(
                        r.status, "Download failed"
                    )


class Http:
    def __init__(self, fs: filesystem.Filesystem) -> None:
        self.fs = fs
        self.session = requests.Session()

    def __enter__(self) -> "Http":
        return self

    def __exit__(self, *args: typing.Any) -> None:
        self.session.close()

    def download_file(
        self,
        url: str,
        path: str,
        headers: typing.Optional[dict[str, str]] = None,
        cookies: typing.Optional[dict[str, str]] = None,
        auth: typing.Optional[requests.auth.HTTPBasicAuth] = None,
    ) -> None:
        self.fs.makedirs(os.path.dirname(path), exist_ok=True)
        with tracer.trace("download_file_task"):
            # TODO: Error handling when no file or manifest

            # yarl.URL encoded=True is very important! Otherwise aiohttp
            # will encode the url again and we'll get a 404 for things like
            # signed URLs
            with self.session.get(
                str(yarl.URL(url, encoded=True)),
                headers=headers,
                cookies=cookies,
                auth=auth,
            ) as r:
                if r.status_code == 200:  # type: ignore
                    with self.fs.open_write(path, mode="wb") as f:
                        f.write(r.content)  # type: ignore
                else:
                    raise server_error_handling.WeaveInternalHttpException.from_code(
                        r.status_code,
                        "Download failed",  # type: ignore
                    )



import typing
from typing import Mapping

from weave.trace.constants import TRACE_CALL_EMOJI

if typing.TYPE_CHECKING:
    from .op_def import OpDef


def print_run_link(run):
    print(f"{TRACE_CALL_EMOJI} {run.ui_url}")


def execute_op(op_def: "OpDef", inputs: Mapping[str, typing.Any]):
    res = op_def.resolve_fn(**inputs)

    return res



from weave.legacy.weave import context_state as _context

_loading_builtins_token = _context.set_loading_built_ins()

from weave.legacy.weave.ecosystem.huggingface.hfmodel import ModelOutputAttention
from weave.legacy.weave.ecosystem.huggingface.huggingface_datasets import *
from weave.legacy.weave.ecosystem.huggingface.huggingface_models import *
from weave.legacy.weave.ecosystem.huggingface.model_textclassification import (
    FullTextClassificationPipelineOutput,
)

_context.clear_loading_built_ins(_loading_builtins_token)



import os

import pytest

from weave.legacy.weave import (
    api,
    context_state,
    mappers_python,
    ops,
    weave_internal,
    weavejs_fixes,
)
from weave.legacy.weave import weave_types as types


@pytest.mark.skip(
    "calling custom ops with graph engine is broken right now. Not needed for weaveflow"
)
def test_clean_opcall_str():
    _loading_builtins_token = context_state.set_loading_built_ins(False)

    try:

        @api.op(input_type={"x": types.Number()}, output_type=types.Number())
        def test_inc(x):
            return x + 1

    finally:
        context_state.clear_loading_built_ins(_loading_builtins_token)

    n = test_inc(5)
    assert "local-artifact://" in n.from_op.name
    fixed_n = weavejs_fixes.remove_opcall_versions_node(n)
    assert fixed_n.from_op.name == "op-test_inc"


def test_convert_specific_op_to_generic_op_node():
    node = ops.local_path(os.path.join("testdata", "cereal.csv")).readcsv()[0]["type"]
    node = weavejs_fixes.fixup_node(node)
    assert node.from_op.name == "pick"
    assert "self" not in node.from_op.inputs
    # Convert inputs to list to ensure we didn't screw up order
    assert list(node.from_op.inputs.keys())[0] == "obj"
    assert "obj" in node.from_op.inputs
    dict_node = node.from_op.inputs["obj"]
    assert dict_node.from_op.name == "index"
    assert "self" not in dict_node.from_op.inputs
    assert list(dict_node.from_op.inputs.keys())[0] == "arr"


def test_convert_specific_op_to_generic_op_data():
    node = ops.local_path(os.path.join("testdata", "cereal.csv")).readcsv()[0]["type"]
    node = weavejs_fixes.fixup_data(node.to_json())
    assert node["fromOp"]["name"] == "pick"
    assert "self" not in node["fromOp"]["inputs"]
    assert list(node["fromOp"]["inputs"].keys())[0] == "obj"
    dict_node = node["fromOp"]["inputs"]["obj"]
    assert dict_node["fromOp"]["name"] == "index"
    assert "self" not in dict_node["fromOp"]["inputs"]
    assert list(dict_node["fromOp"]["inputs"].keys())[0] == "arr"


def test_strip_union_encoding_from_weavejs_response(weave_test_client):
    obj = [1, "a"]
    obj_type = types.TypeRegistry.type_of(obj)
    m = mappers_python.map_to_python(obj_type, None)
    serialized = m.apply(obj)

    assert serialized != obj
    assert all("_union_id" in item for item in serialized)

    node = weave_internal.make_const_node(obj_type, obj)
    serialized2 = weave_test_client.execute([node])[0]

    assert serialized2 == obj



import dataclasses
import pickle
import typing

import huggingface_hub
import torch
import transformers

import weave
from weave.legacy.weave.ecosystem import pytorch
from weave.legacy.weave.ecosystem.huggingface import (
    hfmodel,
    model_textclassification,
    model_textgeneration,
)


def full_model_info_to_hfmodel(
    info: huggingface_hub.hf_api.ModelInfo,
) -> hfmodel.HFModel:
    kwargs = {
        "_id": info.modelId,
        "_sha": info.sha,
        "_pipeline_tag": info.pipeline_tag,
        "_tags": info.tags,
        "_downloads": getattr(info, "downloads", 0),
        "_likes": getattr(info, "likes", 0),
        "_library_name": getattr(info, "library_name", None),
    }
    if info.pipeline_tag == "text-classification":
        return model_textclassification.HFModelTextClassification(**kwargs)
    elif info.pipeline_tag == "text-generation":
        return model_textgeneration.HFModelTextGeneration(**kwargs)
    return hfmodel.HFModel(**kwargs)


@weave.type()
class HuggingfaceModelsPanel(weave.Panel):
    id = "HuggingfaceModelsPanel"
    input_node: weave.Node[list[hfmodel.HFModel]]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Table:
        return weave.legacy.weave.panels.Table(
            self.input_node,
            columns=[
                lambda model_row: weave.legacy.weave.panels.WeaveLink(
                    model_row.id(),
                    to=lambda input: huggingface().model(input),  # type: ignore
                ),
                lambda model_row: model_row.sha(),
                lambda model_row: model_row.pipeline_tag(),
                lambda model_row: model_row.tags(),
                lambda model_row: model_row.downloads(),
                lambda model_row: model_row.likes(),
                lambda model_row: model_row.library_name(),
            ],
        )


@weave.type()
class HuggingfaceModelPanel(weave.Panel):
    id = "HuggingfaceModelPanel"
    input_node: weave.Node[hfmodel.HFModel]

    @weave.op(pure=False)
    def render(self) -> weave.legacy.weave.panels.Card:
        model = typing.cast(hfmodel.HFModel, self.input_node)
        return weave.legacy.weave.panels.Card(
            title=model.id(),
            subtitle="HuggingFace Hub Model",
            content=[
                weave.legacy.weave.panels.CardTab(
                    name="Model Card",
                    content=weave.legacy.weave.panels.PanelMarkdown(model.readme()),  # type: ignore
                ),
                weave.legacy.weave.panels.CardTab(
                    name="Metadata",
                    content=weave.legacy.weave.panels.Group(
                        items={
                            "id": weave.legacy.weave.panels.LabeledItem(
                                item=model.id(), label="ID"
                            ),
                            "pipeline_tag": weave.legacy.weave.panels.LabeledItem(
                                item=model.pipeline_tag(), label="Pipeline tag"
                            ),
                        }
                    ),
                ),
                # Broke in panel refactor. Don't have concrete op name available here so
                # can't get the right type for the output.
                # weave.legacy.weave.panels.CardTab(
                #     name="Inference Logs",
                #     content=weave.legacy.weave.panels.Table(
                #         weave.legacy.weave.ops.used_by(model, model.call.op_name()),
                #         columns=[
                #             lambda run: run.output.model_input,
                #             lambda run: run.output.model_output[0]["generated_text"],
                #         ],
                #     ),
                # ),
            ],
        )


@weave.type()
class HuggingFacePackage:
    @weave.op(hidden=True)
    def model_refine_output_type(self, id: str) -> weave.types.Type:
        api = huggingface_hub.HfApi()
        info = api.model_info(id)
        if info.pipeline_tag == "text-classification":
            return model_textclassification.HFModelTextClassificationType()
        if info.pipeline_tag == "text-generation":
            return model_textgeneration.HFModelTextGenerationType()
        raise Exception(
            "Huggingface model type '%s' not yet supported. Add support in ecosystem/huggingface."
            % info.pipeline_tag
        )

    @weave.op(refine_output_type=model_refine_output_type)
    def model(self, id: str) -> hfmodel.HFModel:
        api = huggingface_hub.HfApi()
        info = api.model_info(id)
        return full_model_info_to_hfmodel(info)

    @weave.op()
    def models(self) -> list[hfmodel.HFModel]:
        api = huggingface_hub.HfApi()
        return [full_model_info_to_hfmodel(info) for info in api.list_models(full=True)]


@weave.op(render_info={"type": "function"})
def huggingface() -> HuggingFacePackage:
    return HuggingFacePackage()


@weave.type()
class HuggingfacePackagePanel(weave.Panel):
    id = "HuggingfacePackagePanel"
    input_node: weave.Node[HuggingFacePackage]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Card:
        pack = typing.cast(HuggingFacePackage, self.input_node)  # type: ignore
        return weave.legacy.weave.panels.Card(
            title="Huggingface Package",
            subtitle="Browse Models and Datasets",
            content=[
                weave.legacy.weave.panels.CardTab(name="Models", content=pack.models()),  # type: ignore
            ],
        )



from weave.legacy.weave import api as weave
from weave.legacy.weave import storage
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.ops_domain import wb_domain_types as wdt


def test_with_keys_assignability():
    org_type = wdt.OrgType
    org_with_keys = org_type.with_keys({"name": types.String()})
    assert not org_with_keys.assign_type(org_type)
    assert org_type.assign_type(org_with_keys)
    assert org_with_keys.assign_type(org_with_keys)

    org_with_keys_2 = org_type.with_keys(
        {
            "name": types.String(),
            "id": types.String(),
        }
    )

    assert org_with_keys.assign_type(org_with_keys_2)


def test_with_keys_not_assignability():
    org_type = wdt.OrgType
    org_with_keys = org_type.with_keys({"name": types.String()})
    project_type = wdt.ProjectType
    project_with_keys = project_type.with_keys({"name": types.String()})

    assert not org_with_keys.assign_type(project_with_keys)
    assert not project_with_keys.assign_type(org_with_keys)


def test_type_of_run_with_keys():
    run = wdt.Run.from_keys({"a": "1"})
    assert types.TypeRegistry.type_of(run) == wdt.RunType.with_keys(
        {"a": types.String()}
    )


def test_serialize_deserialize_run_type():
    run_type = wdt.RunType.with_keys({"a": types.String()})
    assert run_type.__class__.from_dict(run_type.to_dict()) == run_type


def test_storage_on_type_with_keys():
    run = wdt.Run.from_keys({"a": "1"})
    obj_id = storage.save(run, "my-test-run")
    loaded = storage.get(obj_id)
    assert loaded == run


def test_type_of_run_node_with_keys():
    run = wdt.Run.from_keys({"a": "1"})
    node = weave.save(run)
    assert node.type == wdt.RunType.with_keys({"a": types.String()})



import typing

import weave
from weave.legacy.weave import context_state, weave_internal

_loading_builtins_token = context_state.set_loading_built_ins()

MetricRows = typing.TypeVar("MetricRows")


@weave.type()
class _TestType(typing.Generic[MetricRows]):
    metrics: MetricRows


context_state.clear_loading_built_ins(_loading_builtins_token)


def test_const_type_mapper():
    node = weave_internal.make_const_node(
        _TestType.WeaveType(weave.types.List(weave.types.Number())), None
    )
    fn_node = weave_internal.define_fn(
        {"row": weave.types.Number()}, lambda row: row + 1
    )
    node.metrics.map(fn_node)



import dataclasses
import datetime
import hashlib
import json
import logging
import typing
import uuid
from enum import Enum

import typeguard
from wandb.sdk.data_types.trace_tree import Result as WBSpanResult
from wandb.sdk.data_types.trace_tree import Span as WBSpan

from weave.legacy.weave import api as weave
from weave.legacy.weave import stream_data_interfaces
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import op_def
from weave.legacy.weave.decorator_op import op


class StatusCode:
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


class SpanKind:
    LLM = "LLM"
    CHAIN = "CHAIN"
    AGENT = "AGENT"
    TOOL = "TOOL"


@weave.type()
class Result:
    # NOTE: In principle this type should be typing.Optional[typing.Dict[str, typing.Any]],
    # but due to a bug in our langchain integration, we have some cases where
    # inputs was logged as a list[str] instead of a dict[str, Any]. The more flexible type
    # below allows our PanelTraceViewer to work in both cases, but were it not for that bug,
    # this extra flexible type wouldn't be needed.
    inputs: typing.Optional[typing.Union[typing.Dict[str, typing.Any], list[str]]]
    outputs: typing.Optional[typing.Dict[str, typing.Any]]


def _setattr_with_typeguard(obj: typing.Any, key: str, value: typing.Any) -> None:
    """
    Set an attribute on an instance of a class with annotated class attributes, but first check that
    the value is of the correct type. If not, log a warning and set the attribute to None.
    """

    hints = typing.get_type_hints(obj.__class__)
    if key not in hints:
        raise ValueError(f"Object {obj} has no attribute {key}")

    try:
        typeguard.check_type(value, hints[key])
    except typeguard.TypeCheckError as e:
        # warn
        logging.warning(
            f"Setting attribute {key} of {obj} to {value} failed typeguard check: {e}. Replacing with None."
        )
        value = None

    setattr(obj, key, value)


@weave.type()
class Span:
    span_id: typing.Optional[str] = None
    name: typing.Optional[str] = None
    start_time_ms: typing.Optional[int] = None
    end_time_ms: typing.Optional[int] = None
    status_code: typing.Optional[str] = None
    status_message: typing.Optional[str] = None
    attributes: typing.Optional[typing.Dict[str, typing.Any]] = None
    # results is not standard and not representation by OpenTelemetry
    results: typing.Optional[typing.List[typing.Optional[Result]]] = dataclasses.field(
        default_factory=lambda: None
    )
    child_spans: typing.Optional[typing.List[dict]] = dataclasses.field(
        default_factory=lambda: None
    )
    span_kind: typing.Optional[str] = dataclasses.field(default_factory=lambda: None)

    def get_child_spans(self) -> typing.List["Span"]:
        return [Span.from_dump(s) for s in self.child_spans or []]

    @classmethod
    def from_dump(cls, dump_dict: dict) -> "Span":
        root_span = cls()
        for key in dump_dict:
            if key == "results":
                results = dump_dict[key]
                _setattr_with_typeguard(
                    root_span,
                    key,
                    (
                        [Result(**r) if r is not None else None for r in results]
                        if results is not None
                        else None
                    ),
                )
            else:
                _setattr_with_typeguard(root_span, key, dump_dict[key])

        return root_span


# Type used when logging normalized (flat) span data instead of a tree.
# we can add status_code etc...
class FlatSpanType(typing.TypedDict):
    trace_id: typing.Optional[str]
    span_id: typing.Optional[str]
    parent_id: typing.Optional[str]
    name: typing.Optional[str]
    start_time_ms: typing.Optional[int]
    end_time_ms: typing.Optional[int]
    attributes: typing.Optional[typing.Dict[str, typing.Any]]


def stringified_output(obj: typing.Any) -> str:
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, (int, float)):
        return str(obj)
    elif isinstance(obj, dict):
        return json.dumps(obj, indent=2)
    elif isinstance(obj, list):
        return json.dumps(obj, indent=2)
    else:
        return str(obj)


def get_first_error(span: Span) -> typing.Optional[str]:
    if span.status_code == "ERROR":
        return span.status_message or ""
    for child in span.get_child_spans() or []:
        error = get_first_error(child)  # type: ignore
        if error is not None:
            return error
    return None


def standarize_result_inputs(result: typing.Optional[Result]) -> typing.Dict[str, str]:
    if result is None:
        return {}
    if isinstance(result.inputs, dict):
        return result.inputs
    if result.inputs is None:
        return {}
    if isinstance(result.inputs, list):
        # NOTE: In principle this block should not be needed, but due to a bug in our
        # langchain integration, we have some cases where inputs was logged as a list[str]
        # instead of a dict[str, Any]. This block allows PanelTraceViewer to work in both cases,
        # but were it not for that bug, this block wouldn't be needed.
        return {str(i): v for i, v in enumerate(result.inputs)}
    raise ValueError(f"Unexpected result inputs type: {type(result.inputs)}")


def get_trace_input_str(span: Span) -> str:
    return "\n\n".join(
        [
            "\n\n".join(
                [
                    f"**{ndx}.{eKey}:** {eValue}"
                    for eKey, eValue in standarize_result_inputs(result).items()
                ]
            )
            for ndx, result in enumerate(span.results or [])
        ]
    )


def get_trace_output_str(span: Span) -> str:
    return "\n\n".join(
        [
            "\n\n".join(
                [
                    f"**{ndx}.{eKey}:** {eValue}"
                    for eKey, eValue in (
                        (result.outputs or {}) if result is not None else {}
                    ).items()
                ]
            )
            for ndx, result in enumerate(span.results or [])
        ]
    )


def get_chain_repr(span: Span) -> str:
    basic_name: str = span.name or span.span_kind or "Unknown"
    inner_calls = []
    for child in span.get_child_spans():
        inner_calls.append(get_chain_repr(child))  # type: ignore
    if len(inner_calls) == 0:
        return basic_name
    return f"{basic_name}({', '.join(inner_calls)})"


@weave.type(__override_name="wb_trace_tree")  # type: ignore
class WBTraceTree:
    root_span_dumps: str  # Span
    model_dict_dumps: typing.Optional[
        str
    ] = None  # typing.Optional[typing.Dict[str, typing.Any]]
    model_hash: typing.Optional[str] = None

    @weave.op()
    def startTime(self) -> typing.Optional[int]:
        try:
            root_span: Span = json.loads(self.root_span_dumps)
            return root_span.start_time_ms
        except Exception:
            logging.warning("Failed to parse root span")
            return None

    @weave.op()
    def traceSummaryDict(self) -> dict:
        root_span = Span()
        try:
            loaded_dump = json.loads(self.root_span_dumps)
        except Exception:
            logging.warning("Failed to parse root span")
        else:
            root_span = Span.from_dump(loaded_dump)

        return {
            "isSuccess": root_span.status_code in [None, "SUCCESS"],
            "startTime": root_span.start_time_ms,
            "formattedInput": get_trace_input_str(root_span),
            "formattedOutput": get_trace_output_str(root_span),
            "formattedChain": get_chain_repr(root_span),
            "error": get_first_error(root_span),
            "modelHash": self.model_hash,
        }


def span_dict_to_wb_span(span_dict: dict) -> WBSpan:
    child_spans = [
        span_dict_to_wb_span(child_dict)
        for child_dict in (span_dict.get("child_spans") or [])
    ]
    return WBSpan(
        span_id=span_dict.get("span_id"),
        name=span_dict.get("name"),
        start_time_ms=span_dict.get("start_time_ms"),
        end_time_ms=span_dict.get("end_time_ms"),
        status_code=span_dict.get("status_code"),
        status_message=span_dict.get("status_message"),
        attributes=span_dict.get("attributes"),
        results=[
            WBSpanResult(
                inputs=r.get("inputs"),
                outputs=r.get("outputs"),
            )
            for r in span_dict.get("results", [])
        ],
        span_kind=span_dict.get("span_kind"),
        child_spans=child_spans,
    )


class TraceSpanDictWithTimestamp(stream_data_interfaces.TraceSpanDict):
    timestamp: datetime.datetime


@op(
    hidden=True,
)
def refine_convert_output_type(
    tree: WBTraceTree,
) -> types.Type:
    with op_def.no_refine():
        node = convert_to_spans(tree)
    res = weave.use(node)
    if len(res) == 0:
        return types.List(
            types.TypedDict(
                {
                    "span_id": types.String(),
                    "name": types.String(),
                    "trace_id": types.String(),
                    "status_code": types.String(),
                    "start_time_s": types.Number(),
                    "end_time_s": types.Number(),
                    "parent_id": types.optional(types.String()),
                    "attributes": types.optional(types.TypedDict({})),
                    "inputs": types.optional(types.TypedDict({})),
                    "output": types.optional(types.TypedDict({})),
                    "summary": types.optional(types.TypedDict({})),
                    "exception": types.optional(types.String()),
                    "timestamp": types.Timestamp(),
                }
            )
        )
    final = types.TypeRegistry.type_of(res)
    return final


def create_id_from_seed(seed: str) -> str:
    m = hashlib.md5()
    m.update(seed.encode("utf-8"))
    return str(uuid.UUID(m.hexdigest()))


@weave.op(
    name="wb_trace_tree-convertToSpans", refine_output_type=refine_convert_output_type
)
def convert_to_spans(
    tree: WBTraceTree,
) -> typing.List[TraceSpanDictWithTimestamp]:
    loaded_dump = json.loads(tree.root_span_dumps)
    wb_span = span_dict_to_wb_span(loaded_dump)

    # Ensure stable span id (since some old traces don't have them)
    if wb_span.span_id is None:
        wb_span.span_id = create_id_from_seed(tree.root_span_dumps)

    spans: typing.List[
        TraceSpanDictWithTimestamp
    ] = stream_data_interfaces.wb_span_to_weave_spans(
        wb_span, None, None
    )  # type: ignore
    if len(spans) > 0:
        spans[0]["attributes"] = spans[0]["attributes"] or {}
        spans[0]["attributes"]["model"] = {  # type: ignore
            "id": tree.model_hash,
            "obj": tree.model_dict_dumps,
        }

    for span in spans:
        span["timestamp"] = datetime.datetime.fromtimestamp(span["start_time_s"])
    return spans



import re

from weave.legacy.weave import api as weave
from weave.legacy.weave import graph, storage
from weave.legacy.weave.weave_internal import make_const_node

from ...legacy.weave import trace_legacy


def test_node_expr():
    nine = make_const_node(weave.types.Number(), 9)
    res = (nine + 3) * 4
    assert weave.use(res) == 48
    exp = graph.node_expr_str(res)
    # TODO: make node_expr_str use binary ops (make match frontend
    # impl)
    assert exp == "add(9, 3).mult(4)"


# def test_value_expr():
#     nine = make_const_node(weave.types.Number(), 9)
#     res = weave.use((nine + 3) * 4)
#     exp = storage.get_obj_expr(res)
#     assert exp == "add(9, 3).mult(4)"


def test_trace():
    nine = make_const_node(weave.types.Number(), 9)
    res = weave.use((nine + 3) * 4)
    assert res == 48
    mult_run = trace_legacy.get_obj_creator(storage._get_ref(res))
    assert mult_run.op_name == "number-mult"
    assert re.match(
        "^local-artifact://.*run-number-add-.*-output:.*$", str(mult_run.inputs["lhs"])
    )
    assert mult_run.inputs["rhs"] == 4
    add_run = trace_legacy.get_obj_creator(mult_run.inputs["lhs"])
    assert add_run.op_name == "number-add"
    assert add_run.inputs == {"lhs": 9, "rhs": 3}



import dataclasses
import json
import time
import typing

import faiss
from langchain.base_language import BaseLanguageModel
from langchain.callbacks.tracers.base import BaseTracer
from langchain.callbacks.tracers.schemas import Run
from langchain.chains import HypotheticalDocumentEmbedder, RetrievalQA
from langchain.chains.base import Chain
from langchain.chains.combine_documents.base import BaseCombineDocumentsChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains.llm import LLMChain
from langchain.chains.retrieval_qa.base import BaseRetrievalQA
from langchain.chat_models import ChatAnthropic, ChatOpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.docstore.document import Document
from langchain.embeddings.base import Embeddings
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.llms import OpenAI
from langchain.llms.base import BaseLLM
from langchain.llms.openai import BaseOpenAI
from langchain.prompts import (
    BaseChatPromptTemplate,
    BasePromptTemplate,
    ChatPromptTemplate,
    PromptTemplate,
)
from langchain.prompts.base import StringPromptTemplate
from langchain.prompts.chat import (
    AIMessagePromptTemplate,
    BaseMessagePromptTemplate,
    BaseStringMessagePromptTemplate,
    ChatMessagePromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain.schema import BaseRetriever
from langchain.vectorstores import FAISS, VectorStore
from langchain.vectorstores.base import VectorStoreRetriever

import weave
from weave.legacy.weave import storage
from weave.legacy.weave import artifact_base
from weave.legacy.weave.artifact_fs import FilesystemArtifact
from weave.legacy.weave.ecosystem.langchain import util
from weave.legacy.weave.ops_domain import trace_tree
from weave.legacy.weave.weave_types import Type


class WeaveTracer(BaseTracer):
    run: typing.Optional[Run]

    def _persist_run(self, run: "Run") -> None:
        self.run = run


@dataclasses.dataclass(frozen=True)
class DocumentType(weave.types.ObjectType):
    instance_classes = Document

    metadata: weave.types.Type = weave.types.TypedDict()

    def property_types(self) -> dict[str, Type]:
        return {"page_content": weave.types.String(), "metadata": self.metadata}


class VectorStoreType(weave.types.Type):
    instance_classes = VectorStore


# TODO: LangChain uses pydantic, we can infer object types everywhere...
class EmbeddingsType(weave.types.ObjectType):
    instance_classes = Embeddings


class OpenAIEmbeddingsType(EmbeddingsType):
    instance_classes = OpenAIEmbeddings


@dataclasses.dataclass(frozen=True)
class BaseRetrieverType(weave.types.ObjectType):
    instance_classes = BaseRetriever


@dataclasses.dataclass(frozen=True)
class VectorStoreRetrieverType(BaseRetrieverType):
    instance_classes = VectorStoreRetriever

    vectorstore: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "vectorstore": self.vectorstore,
            "search_type": weave.types.String(),
            # TODO: wrong, this is variable.
            "search_kwargs": weave.types.TypedDict(),
        }


@dataclasses.dataclass(frozen=True)
class BasePromptTemplateType(weave.types.ObjectType):
    instance_classes = BasePromptTemplate


@dataclasses.dataclass(frozen=True)
class StringPromptTemplateType(BasePromptTemplateType):
    instance_classes = StringPromptTemplate


@dataclasses.dataclass(frozen=True)
class PromptTemplateType(StringPromptTemplateType):
    instance_classes = PromptTemplate

    def property_types(self) -> dict[str, Type]:
        prop_types = super().property_types()
        prop_types.update(
            {
                "input_variables": weave.types.List(weave.types.String()),
                "template": weave.types.String(),
                "template_format": weave.types.String(),
                "validate_template": weave.types.Boolean(),
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class BaseChatPromptTemplateType(BasePromptTemplateType):
    instance_classes = BaseChatPromptTemplate


@dataclasses.dataclass(frozen=True)
class ChatPromptTemplateType(BaseChatPromptTemplateType):
    instance_classes = ChatPromptTemplate

    messages: weave.types.Type = weave.types.Any()

    def property_types(self) -> dict[str, Type]:
        prop_types = super().property_types()
        prop_types.update(
            {
                "input_variables": weave.types.List(weave.types.String()),
                "messages": self.messages,
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class BaseMessagePromptTemplateType(weave.types.ObjectType):
    instance_classes = BaseMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class BaseStringMessagePromptTemplateType(BaseMessagePromptTemplateType):
    instance_classes = BaseStringMessagePromptTemplate

    prompt: weave.types.Type = weave.types.Any()
    additional_kwargs: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "prompt": self.prompt,
            "additional_kwargs": self.additional_kwargs,
        }


@dataclasses.dataclass(frozen=True)
class ChatMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = ChatMessagePromptTemplate

    def property_types(self):
        prop_types = super().property_types()
        prop_types.update(
            {
                "role": weave.types.String(),
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class HumanMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = HumanMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class AIMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = AIMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class SystemMessagePromptTemplateType(BaseStringMessagePromptTemplateType):
    instance_classes = SystemMessagePromptTemplate


@dataclasses.dataclass(frozen=True)
class ChainType(weave.types.ObjectType):
    instance_classes = Chain


# Hmm... this uses multiple inheritance in langchain, its both ChainType
# and EmbeddingsType. Weave doesn't support that, prefering type dispatch.
@dataclasses.dataclass(frozen=True)
class HyptheticalDocumentEmbedderType(EmbeddingsType):
    instance_classes = HypotheticalDocumentEmbedder

    base_embeddings: weave.types.Type = weave.types.Any()
    llm_chain: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "base_embeddings": self.base_embeddings,
            "llm_chain": self.llm_chain,
        }


@dataclasses.dataclass(frozen=True)
class LLMChainType(ChainType):
    instance_classes = LLMChain

    prompt: weave.types.Type = weave.types.Any()
    llm: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "prompt": self.prompt,
            "llm": self.llm,
            "output_key": weave.types.String(),
        }


@dataclasses.dataclass(frozen=True)
class BaseLanguageModelType(ChainType):
    instance_classes = BaseLanguageModel


@dataclasses.dataclass(frozen=True)
class BaseLLMType(BaseLanguageModelType):
    instance_classes = BaseLLM


@dataclasses.dataclass(frozen=True)
class BaseOpenAIType(BaseLLMType):
    instance_classes = BaseOpenAI

    def property_types(self) -> dict[str, Type]:
        return {
            "model_name": weave.types.String(),
            "temperature": weave.types.Float(),
        }


@dataclasses.dataclass(frozen=True)
class OpenAIType(BaseOpenAIType):
    instance_classes = OpenAI


@dataclasses.dataclass(frozen=True)
class BaseChatModelType(BaseLanguageModelType):
    instance_classes = BaseChatModel


@dataclasses.dataclass(frozen=True)
class ChatOpenAIType(BaseChatModelType):
    instance_classes = ChatOpenAI

    def property_types(self) -> dict[str, Type]:
        return {
            "model_name": weave.types.String(),
            "temperature": weave.types.Float(),
        }


@dataclasses.dataclass(frozen=True)
class ChatAnthropicType(BaseChatModelType):
    instance_classes = ChatAnthropic

    def property_types(self) -> dict[str, Type]:
        return {
            "model": weave.types.String(),
            "temperature": weave.types.Float(),
        }


@dataclasses.dataclass(frozen=True)
class BaseCombineDocumentsChainType(ChainType):
    instance_classes = BaseCombineDocumentsChain

    def property_types(self):
        return {
            "input_key": weave.types.String(),
            "output_key": weave.types.String(),
        }


@dataclasses.dataclass(frozen=True)
class StuffDocumentsChainType(BaseCombineDocumentsChainType):
    instance_classes = StuffDocumentsChain

    llm_chain: weave.types.Type = weave.types.Any()
    document_prompt: weave.types.Type = weave.types.Any()

    def property_types(self):
        prop_types = super().property_types()
        prop_types.update(
            {
                "llm_chain": self.llm_chain,
                "document_prompt": self.document_prompt,
                "document_variable_name": weave.types.String(),
                "document_separator": weave.types.String(),
            }
        )
        return prop_types


@dataclasses.dataclass(frozen=True)
class BaseRetrievalQAType(ChainType):
    instance_classes = BaseRetrievalQA

    combine_documents_chain: weave.types.Type = weave.types.Any()

    def property_types(self):
        return {
            "combine_documents_chain": self.combine_documents_chain,
            "input_key": weave.types.String(),
            "output_key": weave.types.String(),
            "return_source_documents": weave.types.Boolean(),
        }


@dataclasses.dataclass(frozen=True)
class RetrievalQAType(BaseRetrievalQAType):
    instance_classes = RetrievalQA

    retriever: weave.types.Type = weave.types.Any()

    def property_types(self):
        prop_types = super().property_types()
        prop_types.update(
            {
                "retriever": self.retriever,
            }
        )
        return prop_types


class FAISSType(VectorStoreType):
    instance_classes = FAISS

    def save_instance(
        self, obj: FAISS, artifact, name
    ) -> typing.Union[list[str], artifact_base.ArtifactRef, None]:
        # Langchain is inefficient, it resaves all the documents.
        # But in weave we've probably already saved them, so we'd prefer to just store a ref.
        # TODO: replace with weave implementation
        with artifact.new_dir(name) as dir:
            obj.save_local(dir)
        return None

    def load_instance(
        self,
        artifact: FilesystemArtifact,
        name: str,
        extra: typing.Union[list[str], None] = None,
    ) -> typing.Any:
        # TODO: LangChain requires Embeddings when loading, but it drops the embeddings
        # object!
        return FAISS.load_local(artifact.path(name), OpenAIEmbeddings())


@weave.op()
def faiss_from_documents(
    documents: list[Document],
    embeddings: Embeddings,
) -> FAISS:
    return FAISS.from_documents(documents, embeddings)


class DocumentEmbedding(typing.TypedDict):
    document: Document
    # TODO: np array
    embedding: list[float]
    # I made this a string because there's no way to tell Plot
    # its categorical at the moment.
    cluster: str


@weave.weave_class(weave_type=FAISSType)
class FaissOps:
    @weave.op()
    def document_embeddings(vector_store: FAISS) -> list[DocumentEmbedding]:
        embeddings = vector_store.index.reconstruct_n()
        kmeans = faiss.Kmeans(embeddings.shape[1], 20, niter=20, verbose=False)
        kmeans.train(embeddings)
        cluster_distances, cluster_ids = kmeans.index.search(embeddings, 1)
        cluster_ids = cluster_ids.tolist()
        result: list[DocumentEmbedding] = []
        for i, embedding in enumerate(embeddings):
            result.append(
                {
                    "document": vector_store.docstore.search(
                        vector_store.index_to_docstore_id[i]
                    ),
                    "embedding": embedding.tolist(),
                    "cluster": str(cluster_ids[i][0]),
                }
            )
        return result


@weave.op()
def similarity_search(vector_store: VectorStore, query: str) -> list[Document]:
    return vector_store.similarity_search(query)


@weave.op(render_info={"type": "function"})
def openai_embeddings() -> OpenAIEmbeddings:
    return OpenAIEmbeddings()


@weave.op(render_info={"type": "function"})
def openai(model_name: str, temperature: float) -> OpenAI:
    return OpenAI(model_name=model_name, temperature=temperature)


@weave.op(render_info={"type": "function"})
def chat_openai(model_name: str, temperature: float) -> ChatOpenAI:
    return ChatOpenAI(model_name=model_name, temperature=temperature)


@weave.op(render_info={"type": "function"})
def chat_anthropic(model_name: str, temperature: float) -> ChatAnthropic:
    return ChatAnthropic(model=model_name, temperature=temperature)


@weave.op(render_info={"type": "function"})
def llm_chain(llm: BaseLanguageModel, prompt: BasePromptTemplate) -> LLMChain:
    return LLMChain(llm=llm, prompt=prompt)


@weave.op(render_info={"type": "function"})
def hypothetical_document_embedder(
    base_embeddings: Embeddings, llm_chain: LLMChain
) -> HypotheticalDocumentEmbedder:
    return HypotheticalDocumentEmbedder(
        base_embeddings=base_embeddings, llm_chain=llm_chain
    )


@weave.op(render_info={"type": "function"})
def retrieval_qa_from_chain_type(
    language_model: BaseLanguageModel, chain_type: str, retriever: VectorStore
) -> RetrievalQA:
    retriever = retriever.as_retriever()
    return RetrievalQA.from_chain_type(
        llm=language_model,
        chain_type=chain_type,
        retriever=retriever,
    )


@weave.weave_class(weave_type=BaseChatModelType)
class BaseChatModelOps:
    @weave.op()
    def predict(self: BaseChatModel, text: str) -> str:
        if text == None:
            # TODO: weave engine doesn't handle nullability on args that aren't the first
            return None  # type: ignore
        return self.predict(text)


@weave.weave_class(weave_type=BaseLLMType)
class BaseLLMOps:
    @weave.op()
    def predict(self: BaseLLM, text: str) -> str:
        if text == None:
            # TODO: weave engine doesn't handle nullability on args that aren't the first
            return None  # type: ignore
        return self.predict(text)


ChainTypeVar = typing.TypeVar("ChainTypeVar", bound=Chain)


@weave.type()
class ChainRunResult(typing.Generic[ChainTypeVar]):
    chain: ChainTypeVar
    query: str
    result: str
    latency: float
    trace: trace_tree.WBTraceTree


@weave.weave_class(weave_type=ChainType)
class ChainOps:
    @weave.op(
        output_type=lambda input_type: ChainRunResult.WeaveType(  # type: ignore
            chain=input_type["chain"]
        )
    )
    def run(chain: Chain, query: str):
        if query == None:
            # TODO: weave engine doesn't handle nullability on args that aren't the first
            return None  # type: ignore
        tracer = WeaveTracer()
        start_time = time.time()
        result = chain.run(query, callbacks=[tracer])
        latency = time.time() - start_time
        lc_run = tracer.run
        if lc_run is None:
            raise ValueError("LangChain run was not recorded")
        span = util.safely_convert_lc_run_to_wb_span(lc_run)
        if span is None:
            raise ValueError("Could not convert LangChain run to Weave span")

        # Returns the langchain trace as part of the result... not really what we
        # want.
        return ChainRunResult(
            chain=chain,
            query=query,
            result=result,
            latency=latency,
            trace=trace_tree.WBTraceTree(json.dumps(storage.to_python(span)["_val"])),
        )


# @weave.weave_class(weave_type=BaseRetrievalQAType)
# class BaseRetrievalQAOps:
#     @weave.op()
#     def run(chain: BaseRetrievalQA, query: str) -> RunResult:
#         if query == None:
#             # TODO: weave engine doesn't handle nullability on args that aren't the first
#             return None  # type: ignore
#         tracer = WeaveTracer()
#         result = chain.run(query, callbacks=[tracer])
#         lc_run = tracer.run
#         if lc_run is None:
#             raise ValueError("LangChain run was not recorded")
#         span = util.safely_convert_lc_run_to_wb_span(lc_run)
#         if span is None:
#             raise ValueError("Could not convert LangChain run to Weave span")

#         # Returns the langchain trace as part of the result... not really what we
#         # want.
#         return RunResult(
#             result=result,
#             trace=trace_tree.WBTraceTree(json.dumps(storage.to_python(span)["_val"])),
#         )



from weave.legacy.weave import decorator_type


@decorator_type.type(
    "stream_table",
)
class StreamTableType:
    table_name: str  # maps to run name in W&B data model
    project_name: str
    entity_name: str



from weave.legacy.weave import api as weave
from weave.legacy.weave import ops_arrow as arrow
from weave.legacy.weave.ops_primitives import list_


class ListLikeNodeInterface:
    @staticmethod
    def make_node(value):
        raise NotImplementedError

    @staticmethod
    def use_node(node):
        raise NotImplementedError


class ListNode(ListLikeNodeInterface):
    @staticmethod
    def make_node(value):
        return list_.make_list(**{f"{i}": v for i, v in enumerate(value)})

    @staticmethod
    def use_node(node):
        return weave.use(node)


class ArrowNode(ListLikeNodeInterface):
    @staticmethod
    def make_node(value):
        return weave.save(arrow.to_arrow(value))

    @staticmethod
    def use_node(node):
        return weave.use(node).to_pylist_notags()


ListInterfaces = [ListNode, ArrowNode]



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import box
from weave.legacy.weave.language_features.tagging import tagged_value_type

if typing.TYPE_CHECKING:
    from weave.legacy.weave import op_def as OpDef


def should_force_none_result(
    inputs: dict[str, typing.Any], op_def: "OpDef.OpDef"
) -> bool:
    # Hacking... this is nullability of ops
    # We should do this as a compile pass instead of hard-coding in engine.
    # That means we need an op called like "handle_null" that takes a function
    # as its second argument. Function is the op we want to execute if non-null.
    # TODO: fix
    # TODO: not implemented for async ops
    if inputs and not op_def.name.endswith("coalesce"):
        input0 = list(inputs.values())[0]
        named_args = op_def.input_type.named_args()
        return (
            len(named_args) > 0
            and (input0 is None or isinstance(input0, box.BoxedNone))
            and not isinstance(
                op_def.concrete_output_type, tagged_value_type.TaggedValueType
            )
            and not isinstance(named_args[0].type, tagged_value_type.TaggedValueType)
            and not types.is_optional(named_args[0].type)
        )
    return False



import dataclasses
import os
import typing
import logging


ValueType = typing.TypeVar("ValueType")
ValueType2 = typing.TypeVar("ValueType2")

DEBUG = False or os.environ.get("WEAVE_VALUE_OR_ERROR_DEBUG", False)


class _ValueOrErrorInterface(typing.Generic[ValueType]):
    def unwrap(self) -> ValueType:
        raise NotImplementedError

    def transform_and_catch(
        self, fn: typing.Callable[[ValueType], ValueType2]
    ) -> "_ValueOrErrorInterface[ValueType2]":
        raise NotImplementedError

    def apply_and_catch(
        self, fn: typing.Callable[[ValueType], typing.Any]
    ) -> "_ValueOrErrorInterface[ValueType]":
        raise NotImplementedError


@dataclasses.dataclass
class Value(_ValueOrErrorInterface[ValueType]):
    _value: ValueType

    def unwrap(self) -> ValueType:
        return self._value

    def transform_and_catch(
        self, fn: typing.Callable[[ValueType], ValueType2]
    ) -> "ValueOrError[ValueType2]":
        try:
            return Value(fn(self._value))
        except Exception as e:
            return Error(e)

    def apply_and_catch(
        self, fn: typing.Callable[[ValueType], typing.Any]
    ) -> "ValueOrError[ValueType]":
        try:
            fn(self._value)
            return self
        except Exception as e:
            return Error(e)


@dataclasses.dataclass
class Error(_ValueOrErrorInterface[typing.Any]):
    _error: Exception

    def __post_init__(self) -> None:
        if DEBUG:
            raise self._error

    def unwrap(self) -> typing.Any:
        raise self._error

    def transform_and_catch(
        self, fn: typing.Callable[[typing.Any], typing.Any]
    ) -> "ValueOrError[typing.Any]":
        return self

    def apply_and_catch(
        self, fn: typing.Callable[[typing.Any], typing.Any]
    ) -> "ValueOrError[typing.Any]":
        return self


ValueOrError = typing.Union[Value[ValueType], Error]


@dataclasses.dataclass
class ValueOrErrors(typing.Generic[ValueType]):
    _items: list[ValueOrError[ValueType]] = dataclasses.field(default_factory=list)

    @classmethod
    def from_values(cls, values: list[ValueType]) -> "ValueOrErrors[ValueType]":
        return ValueOrErrors([Value(i) for i in values])

    def safe_map(
        self, fn: typing.Callable[[ValueType], ValueType2]
    ) -> "ValueOrErrors[ValueType2]":
        return ValueOrErrors([i.transform_and_catch(fn) for i in self._items])

    def raw_map(
        self,
        fn: typing.Callable[[ValueType], ValueOrError[ValueType2]],
    ) -> "ValueOrErrors[ValueType2]":
        return ValueOrErrors(
            [fn(i._value) if isinstance(i, Value) else i for i in self._items]
        )

    def safe_apply(
        self, fn: typing.Callable[[ValueType], typing.Any]
    ) -> "ValueOrErrors[ValueType]":
        return ValueOrErrors([i.apply_and_catch(fn) for i in self._items])

    def batch_map(
        self, fn: typing.Callable[[list[ValueType]], "ValueOrErrors[ValueType2]"]
    ) -> "ValueOrErrors[ValueType2]":
        valid_inputs = [i._value for i in self._items if isinstance(i, Value)]
        valid_results = fn(valid_inputs)
        valid_index = 0
        res: list[ValueOrError[ValueType2]] = []
        for i in self._items:
            if isinstance(i, Error):
                res.append(i)
            else:
                res.append(valid_results._items[valid_index])
                valid_index += 1
        return ValueOrErrors(res)

    def zip(
        self, other: "ValueOrErrors[ValueType2]"
    ) -> "ValueOrErrors[typing.Tuple[ValueType, ValueType2]]":
        res: list[ValueOrError[typing.Tuple[ValueType, ValueType2]]] = []
        for i, j in zip(self._items, other._items):
            if isinstance(i, Error):
                res.append(i)
            elif isinstance(j, Error):
                res.append(j)
            else:
                res.append(Value((i._value, j._value)))
        return ValueOrErrors(res)

    def __iter__(self) -> typing.Iterator[ValueType]:
        for i in self._items:
            yield i.unwrap()

    def __len__(self) -> int:
        return len(self._items)

    def __getitem__(self, index: int) -> ValueType:
        return self._items[index].unwrap()

    def unwrap(
        self,
    ) -> list[ValueType]:
        return list(self.__iter__())

    def iter_items(
        self,
    ) -> typing.Iterator[typing.Tuple[ValueType, typing.Optional[Exception]]]:
        for i in self._items:
            if isinstance(i, Value):
                yield i._value, None
            else:
                yield None, i._error  # type: ignore



import os
from typing import Generator, List, Tuple

import pytest
import tiktoken

import weave
from weave.integrations.integration_utilities import (
    filter_body,
    flatten_calls,
    op_name_from_ref,
)
from weave.trace.weave_client import Call, WeaveClient
from weave.trace_server import trace_server_interface as tsi


@pytest.fixture(scope="session", autouse=True)
def ensure_tiktoken_file() -> Generator[None, None, None]:
    enc = tiktoken.get_encoding("cl100k_base")
    enc.encode("Test")
    yield


def assert_ends_and_errors(calls: list[tuple[Call, int]]) -> None:
    for call, depth in calls:
        assert call.ended_at is not None
        assert call.exception is None


def assert_correct_calls_for_chain_invoke(
    calls: list[Call],
    prompt_template_name_part: str = "PromptTemplate",
) -> None:
    flattened = flatten_calls(calls)
    assert len(flattened) == 4
    assert_ends_and_errors(flattened)

    got = [(op_name_from_ref(c.op_name), d) for (c, d) in flattened]
    exp = [
        ("langchain.Chain.RunnableSequence", 0),
        ("langchain.Prompt." + prompt_template_name_part, 1),
        ("langchain.Llm.ChatOpenAI", 1),
        ("openai.chat.completions.create", 2),
    ]
    assert got == exp


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_simple_chain_invoke(
    client: WeaveClient,
) -> None:
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")
    long_str = (
        "really_massive_name_that_is_longer_than_max_characters_which_would_be_crazy"
    )
    name = long_str + long_str
    prompt.name = name

    exp_name = "really_massive_name_that_is_longer_than_max_characte_ff6e_at_is_longer_than_max_characters_which_would_be_crazy"

    llm_chain = prompt | llm
    _ = llm_chain.invoke({"number": 2})

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_invoke(calls, exp_name)


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
@pytest.mark.asyncio
@pytest.mark.skip  # TODO: remove this once the langchain issue is fixed
async def test_simple_chain_ainvoke(
    client: WeaveClient,
) -> None:
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")

    llm_chain = prompt | llm
    _ = await llm_chain.ainvoke({"number": 2})

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_invoke(calls)


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_simple_chain_stream(
    client: WeaveClient,
) -> None:
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")

    llm_chain = prompt | llm
    for _ in llm_chain.stream({"number": 2}):
        pass

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_invoke(calls)


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
@pytest.mark.asyncio
@pytest.mark.skip
async def test_simple_chain_astream(
    client: WeaveClient,
) -> None:
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")

    llm_chain = prompt | llm
    async for _ in llm_chain.astream({"number": 2}):
        pass

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_invoke(calls)


def assert_correct_calls_for_chain_batch(calls: list[Call]) -> None:
    flattened = flatten_calls(calls)
    assert len(flattened) == 8
    assert_ends_and_errors(flattened)

    got = [(op_name_from_ref(c.op_name), d, c.parent_id) for (c, d) in flattened]
    ids = [c.id for (c, _) in flattened]

    exp = [
        ("langchain.Chain.RunnableSequence", 0, None),
        ("langchain.Prompt.PromptTemplate", 1, ids[0]),
        ("langchain.Llm.ChatOpenAI", 1, ids[0]),
        ("openai.chat.completions.create", 2, ids[2]),
        ("langchain.Chain.RunnableSequence", 0, None),
        ("langchain.Prompt.PromptTemplate", 1, ids[4]),
        ("langchain.Llm.ChatOpenAI", 1, ids[4]),
        ("openai.chat.completions.create", 2, ids[6]),
    ]
    assert got == exp


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_simple_chain_batch(
    client: WeaveClient,
) -> None:
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")

    llm_chain = prompt | llm
    _ = llm_chain.batch([{"number": 2}, {"number": 3}])

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_batch(calls)


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
@pytest.mark.asyncio
@pytest.mark.skip  # TODO: remove this once the langchain issue is fixed
async def test_simple_chain_abatch(
    client: WeaveClient,
) -> None:
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")

    llm_chain = prompt | llm
    _ = await llm_chain.abatch([{"number": 2}, {"number": 3}])

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_batch(calls)


def assert_correct_calls_for_chain_batch_from_op(calls: list[Call]) -> None:
    flattened = flatten_calls(calls)
    assert len(flattened) == 9
    assert_ends_and_errors(flattened)

    got = [(op_name_from_ref(c.op_name), d, c.parent_id) for (c, d) in flattened]
    ids = [c.id for (c, _) in flattened]

    exp = [
        ("run_batch", 0, None),
        ("langchain.Chain.RunnableSequence", 1, ids[0]),
        ("langchain.Prompt.PromptTemplate", 2, ids[1]),
        ("langchain.Llm.ChatOpenAI", 2, ids[1]),
        ("openai.chat.completions.create", 3, ids[3]),
        ("langchain.Chain.RunnableSequence", 1, ids[0]),
        ("langchain.Prompt.PromptTemplate", 2, ids[5]),
        ("langchain.Llm.ChatOpenAI", 2, ids[5]),
        ("openai.chat.completions.create", 3, ids[7]),
    ]
    assert got == exp


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_simple_chain_batch_inside_op(
    client: WeaveClient,
) -> None:
    # This test is the same as test_simple_chain_batch, but ensures things work when nested in an op
    from langchain_core.prompts import PromptTemplate
    from langchain_openai import ChatOpenAI

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )
    prompt = PromptTemplate.from_template("1 + {number} = ")

    llm_chain = prompt | llm

    @weave.op()
    def run_batch(batch: list) -> None:
        _ = llm_chain.batch(batch)

    run_batch([{"number": 2}, {"number": 3}])

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_chain_batch_from_op(calls)


def assert_correct_calls_for_rag_chain(calls: list[Call]) -> None:
    flattened = flatten_calls(calls)
    assert len(flattened) == 10
    assert_ends_and_errors(flattened)

    got = [(op_name_from_ref(c.op_name), d) for (c, d) in flattened]

    exp = [
        ("langchain.Chain.RunnableSequence", 0),
        ("langchain.Chain.RunnableParallel_context_question", 1),
        ("langchain.Chain.RunnableSequence", 2),
        ("langchain.Retriever.VectorStoreRetriever", 3),
        ("langchain.Chain.format_docs", 3),
        ("langchain.Chain.RunnablePassthrough", 2),
        ("langchain.Prompt.ChatPromptTemplate", 1),
        ("langchain.Llm.ChatOpenAI", 1),
        ("openai.chat.completions.create", 2),
        ("langchain.Parser.StrOutputParser", 1),
    ]
    assert got == exp


@pytest.fixture
def fix_chroma_ci() -> Generator[None, None, None]:
    # According to https://docs.trychroma.com/troubleshooting#sqlite
    # which references https://gist.github.com/defulmere/8b9695e415a44271061cc8e272f3c300,
    # on Linux machines (CI runners), we need to patch sqlite3 to pysqlite3 and ensure
    # pysqlite3 is installed.
    if not os.environ.get("CI"):
        yield None
        return

    __import__("pysqlite3")
    import sys

    old = sys.modules["sqlite3"]
    sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")
    try:
        yield None
    finally:
        sys.modules["sqlite3"] = old


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_simple_rag_chain(client: WeaveClient, fix_chroma_ci: None) -> None:
    from typing import List

    from langchain.text_splitter import RecursiveCharacterTextSplitter
    from langchain_community.document_loaders import TextLoader
    from langchain_community.vectorstores import Chroma
    from langchain_core.documents import Document
    from langchain_core.output_parsers import StrOutputParser
    from langchain_core.prompts import ChatPromptTemplate
    from langchain_core.runnables import RunnablePassthrough
    from langchain_openai import ChatOpenAI, OpenAIEmbeddings

    loader = TextLoader("integrations/langchain/test_data/paul_graham_essay.txt")
    docs = loader.load()

    text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=200)
    splits = text_splitter.split_documents(docs)

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    vectorstore = Chroma.from_documents(
        documents=splits, embedding=OpenAIEmbeddings(openai_api_key=api_key)
    )
    retriever = vectorstore.as_retriever()

    prompt = ChatPromptTemplate.from_template(
        "You are an assistant for question-answering tasks. "
        "Use the following pieces of retrieved context to answer the question. "
        "If you don't know the answer, just say that you don't know. "
        "Use three sentences maximum and keep the answer concise.\n"
        "Question: {question}\nContext: {context}\nAnswer:"
    )

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )

    def format_docs(documents: List[Document]) -> str:
        return "\n\n".join(doc.page_content for doc in documents)

    # Chain
    rag_chain = (
        {"context": retriever | format_docs, "question": RunnablePassthrough()}
        | prompt
        | llm
        | StrOutputParser()
    )

    rag_chain.invoke(
        input="What is the essay about?",
    )

    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_rag_chain(calls)


def assert_correct_calls_for_agent_with_tool(calls: list[Call]) -> None:
    flattened = flatten_calls(calls)
    assert len(flattened) == 10
    assert_ends_and_errors(flattened)

    got = [(op_name_from_ref(c.op_name), d) for (c, d) in flattened]

    exp = [
        ("langchain.Chain.AgentExecutor", 0),
        ("langchain.Chain.RunnableSequence", 1),
        ("langchain.Chain.RunnableParallel_input_chat_history_agent_scratchpad", 2),
        ("langchain.Chain.RunnableLambda", 3),
        ("langchain.Chain.RunnableLambda", 3),
        ("langchain.Chain.RunnableLambda", 3),
        ("langchain.Prompt.ChatPromptTemplate", 2),
        ("langchain.Llm.ChatOpenAI", 2),
        ("openai.chat.completions.create", 3),
        ("langchain.Parser.OpenAIFunctionsAgentOutputParser", 2),
    ]
    assert got == exp


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_agent_run_with_tools(
    client: WeaveClient,
) -> None:
    from langchain.agents import AgentExecutor
    from langchain.agents.format_scratchpad import format_to_openai_function_messages
    from langchain.agents.output_parsers import OpenAIFunctionsAgentOutputParser
    from langchain.pydantic_v1 import BaseModel, Field
    from langchain.tools import StructuredTool
    from langchain_core.messages import AIMessage, HumanMessage
    from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
    from langchain_core.utils.function_calling import convert_to_openai_tool
    from langchain_openai import ChatOpenAI

    class CalculatorInput(BaseModel):
        a: int = Field(description="first number")
        b: int = Field(description="second number")

    def multiply(a: int, b: int) -> int:
        """Multiply two numbers."""
        return a * b

    calculator = StructuredTool.from_function(
        func=multiply,
        name="Calculator",
        description="multiply numbers",
        args_schema=CalculatorInput,
        return_direct=True,
    )

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )

    tools = [calculator]
    functions = [convert_to_openai_tool(t) for t in tools]

    assistant_system_message = """You are a helpful assistant. \
    Use tools (only if necessary) to best answer the users questions."""
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", assistant_system_message),
            MessagesPlaceholder(variable_name="chat_history"),
            ("user", "{input}"),
            MessagesPlaceholder(variable_name="agent_scratchpad"),
        ]
    )

    llm_with_tools = llm.bind(tools=functions)

    def _format_chat_history(chat_history: List[Tuple[str, str]]) -> List:
        buffer = []
        for human, ai in chat_history:
            buffer.append(HumanMessage(content=human))
            buffer.append(AIMessage(content=ai))
        return buffer

    agent = (
        {
            "input": lambda x: x["input"],
            "chat_history": lambda x: _format_chat_history(x["chat_history"]),
            "agent_scratchpad": lambda x: format_to_openai_function_messages(
                x["intermediate_steps"]
            ),
        }
        | prompt
        | llm_with_tools
        | OpenAIFunctionsAgentOutputParser()
    )

    class AgentInput(BaseModel):
        input: str
        chat_history: List[Tuple[str, str]] = Field(
            ...,
            extra={"widget": {"type": "chat", "input": "input", "output": "output"}},
        )

    agent_executor = AgentExecutor(agent=agent, tools=tools).with_types(
        input_type=AgentInput
    )

    _ = agent_executor.invoke(
        {"input": "What is 3 times 4 ?", "chat_history": []},
    )
    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_agent_with_tool(calls)


def assert_correct_calls_for_agent_with_function_call(calls: list[Call]) -> None:
    flattened = flatten_calls(calls)
    assert len(flattened) == 11
    assert_ends_and_errors(flattened)

    got = [(op_name_from_ref(c.op_name), d) for (c, d) in flattened]

    exp = [
        ("langchain.Chain.AgentExecutor", 0),
        ("langchain.Chain.RunnableSequence", 1),
        ("langchain.Chain.RunnableParallel_input_chat_history_agent_scratchpad", 2),
        ("langchain.Chain.RunnableLambda", 3),
        ("langchain.Chain.RunnableLambda", 3),
        ("langchain.Chain.RunnableLambda", 3),
        ("langchain.Prompt.ChatPromptTemplate", 2),
        ("langchain.Llm.ChatOpenAI", 2),
        ("openai.chat.completions.create", 3),
        ("langchain.Parser.OpenAIFunctionsAgentOutputParser", 2),
        ("langchain.Tool.Calculator", 1),
    ]
    assert got == exp


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_agent_run_with_function_call(
    client: WeaveClient,
) -> None:
    from langchain.agents import AgentExecutor
    from langchain.agents.format_scratchpad import format_to_openai_function_messages
    from langchain.agents.output_parsers import OpenAIFunctionsAgentOutputParser
    from langchain.pydantic_v1 import BaseModel, Field
    from langchain.tools import StructuredTool
    from langchain_core.messages import AIMessage, HumanMessage
    from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
    from langchain_core.utils.function_calling import convert_to_openai_function
    from langchain_openai import ChatOpenAI

    class CalculatorInput(BaseModel):
        a: int = Field(description="first number")
        b: int = Field(description="second number")

    def multiply(a: int, b: int) -> int:
        """Multiply two numbers."""
        return a * b

    calculator = StructuredTool.from_function(
        func=multiply,
        name="Calculator",
        description="multiply numbers",
        args_schema=CalculatorInput,
        return_direct=True,
    )

    api_key = os.environ.get("OPENAI_API_KEY", "sk-1234567890abcdef1234567890abcdef")

    llm = ChatOpenAI(
        model_name="gpt-3.5-turbo", openai_api_key=api_key, temperature=0.0
    )

    tools = [calculator]
    functions = [convert_to_openai_function(t) for t in tools]

    assistant_system_message = """You are a helpful assistant. \
        Use tools (only if necessary) to best answer the users questions."""
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", assistant_system_message),
            MessagesPlaceholder(variable_name="chat_history"),
            ("user", "{input}"),
            MessagesPlaceholder(variable_name="agent_scratchpad"),
        ]
    )

    llm_with_tools = llm.bind(functions=functions)

    def _format_chat_history(chat_history: List[Tuple[str, str]]) -> List:
        buffer = []
        for human, ai in chat_history:
            buffer.append(HumanMessage(content=human))
            buffer.append(AIMessage(content=ai))
        return buffer

    agent = (
        {
            "input": lambda x: x["input"],
            "chat_history": lambda x: _format_chat_history(x["chat_history"]),
            "agent_scratchpad": lambda x: format_to_openai_function_messages(
                x["intermediate_steps"]
            ),
        }
        | prompt
        | llm_with_tools
        | OpenAIFunctionsAgentOutputParser()
    )

    class AgentInput(BaseModel):
        input: str
        chat_history: List[Tuple[str, str]] = Field(
            ...,
            extra={"widget": {"type": "chat", "input": "input", "output": "output"}},
        )

    agent_executor = AgentExecutor(agent=agent, tools=tools).with_types(
        input_type=AgentInput
    )

    _ = agent_executor.invoke(
        {"input": "What is 3 times 4 ?", "chat_history": []},
    )
    calls = list(client.calls(filter=tsi.CallsFilter(trace_roots_only=True)))
    assert_correct_calls_for_agent_with_function_call(calls)



from datetime import datetime, timedelta, timezone

import pytest

import weave


def test_timestamp_bins_fixed():
    bin_fn = weave.legacy.weave.ops.timestamp_bins_fixed(60)

    ts = datetime(2020, 1, 1, 8, 30, 1)
    called = bin_fn(ts)

    res = weave.use(called)
    assert res["start"] == datetime(2020, 1, 1, 8, 30, 0)
    assert res["stop"] == datetime(2020, 1, 1, 8, 31, 0)


def test_timestamp_bins_nice():
    start_ts = datetime(2020, 1, 1, 8, 30, 1)
    stop_ts = datetime(2020, 1, 1, 12, 30, 0)
    bin_fn = weave.legacy.weave.ops.timestamp_bins_nice([start_ts, stop_ts], 100)

    ts = datetime(2020, 1, 1, 8, 45, 13)
    called = bin_fn(ts)

    res = weave.use(called)
    assert res["start"] == datetime(2020, 1, 1, 8, 45, 0)
    assert res["stop"] == datetime(2020, 1, 1, 8, 46, 0)


def test_timestamp_bin():
    start_ts = datetime(2020, 1, 1, 8, 30, 1)
    stop_ts = datetime(2020, 1, 1, 12, 30, 0)
    bin_fn = weave.legacy.weave.ops.timestamp_bins_nice([start_ts, stop_ts], 100)

    ts = datetime(2020, 1, 1, 8, 45, 13)
    ts_node = weave.save(ts)

    res2 = weave.use(ts_node.bin(bin_fn))
    assert res2["start"] == datetime(2020, 1, 1, 8, 45, 0)
    assert res2["stop"] == datetime(2020, 1, 1, 8, 46, 0)


def test_timestamp_bin_vector():
    start_ts = datetime(2020, 1, 1, 8, 30, 1)
    stop_ts = datetime(2020, 1, 1, 12, 30, 0)
    bin_fn = weave.legacy.weave.ops.timestamp_bins_nice([start_ts, stop_ts], 100)

    ts = datetime(2020, 1, 1, 8, 45, 13, tzinfo=timezone.utc)
    ts_node = weave.legacy.weave.ops.to_weave_arrow([ts, ts + timedelta(seconds=90)])

    # This does not vectorize because mapped ops don't automatically
    # vectorize!
    bin_called = ts_node.bin(bin_fn)
    res = weave.use(bin_called)
    assert res[0]["start"] == datetime(2020, 1, 1, 8, 45, 0, tzinfo=timezone.utc)
    assert res[0]["stop"] == datetime(2020, 1, 1, 8, 46, 0, tzinfo=timezone.utc)
    assert res[1]["start"] == datetime(2020, 1, 1, 8, 46, 0, tzinfo=timezone.utc)
    assert res[1]["stop"] == datetime(2020, 1, 1, 8, 47, 0, tzinfo=timezone.utc)

    mapped_bin_called = ts_node.map(lambda x: x.bin(bin_fn))
    res2 = weave.use(mapped_bin_called)
    assert res2[0]["start"] == datetime(2020, 1, 1, 8, 45, 0, tzinfo=timezone.utc)
    assert res2[0]["stop"] == datetime(2020, 1, 1, 8, 46, 0, tzinfo=timezone.utc)
    assert res2[1]["start"] == datetime(2020, 1, 1, 8, 46, 0, tzinfo=timezone.utc)
    assert res2[1]["stop"] == datetime(2020, 1, 1, 8, 47, 0, tzinfo=timezone.utc)



import typing
from typing import Any

from weave.trace import custom_objs
from weave.trace.object_record import ObjectRecord
from weave.trace.refs import ObjectRef, TableRef, parse_uri
from weave.trace_server.trace_server_interface import (
    FileContentReadReq,
    FileCreateReq,
    TraceServerInterface,
)


def to_json(obj: Any, project_id: str, server: TraceServerInterface) -> Any:
    if isinstance(obj, TableRef):
        return obj.uri()
    elif isinstance(obj, ObjectRef):
        return obj.uri()
    elif isinstance(obj, ObjectRecord):
        res = {"_type": obj._class_name}
        for k, v in obj.__dict__.items():
            res[k] = to_json(v, project_id, server)
        return res
    elif isinstance_namedtuple(obj):
        return {k: to_json(v, project_id, server) for k, v in obj._asdict().items()}
    elif isinstance(obj, (list, tuple)):
        return [to_json(v, project_id, server) for v in obj]
    elif isinstance(obj, dict):
        return {k: to_json(v, project_id, server) for k, v in obj.items()}

    if isinstance(obj, (int, float, str, bool)) or obj is None:
        return obj

    encoded = custom_objs.encode_custom_obj(obj)
    if encoded is None:
        return fallback_encode(obj)
    file_digests = {}
    for name, val in encoded["files"].items():
        file_response = server.file_create(
            FileCreateReq(project_id=project_id, name=name, content=val)
        )
        file_digests[name] = file_response.digest
    result = {
        "_type": encoded["_type"],
        "weave_type": encoded["weave_type"],
        "files": file_digests,
    }
    load_op_uri = encoded.get("load_op")
    if load_op_uri:
        result["load_op"] = load_op_uri
    return result


REP_LIMIT = 1000


def fallback_encode(obj: Any) -> Any:
    rep = None
    try:
        rep = repr(obj)
    except Exception:
        try:
            rep = str(obj)
        except Exception:
            rep = f"<{type(obj).__name__}: {id(obj)}>"
    if isinstance(rep, str):
        if len(rep) > REP_LIMIT:
            rep = rep[:REP_LIMIT] + "..."
    return rep


def isinstance_namedtuple(obj: Any) -> bool:
    return (
        isinstance(obj, tuple) and hasattr(obj, "_asdict") and hasattr(obj, "_fields")
    )


def _load_custom_obj_files(
    project_id: str, server: TraceServerInterface, file_digests: dict
) -> typing.Dict[str, bytes]:
    loaded_files: typing.Dict[str, bytes] = {}
    for name, digest in file_digests.items():
        file_response = server.file_content_read(
            FileContentReadReq(project_id=project_id, digest=digest)
        )
        loaded_files[name] = file_response.content
    return loaded_files


def from_json(obj: Any, project_id: str, server: TraceServerInterface) -> Any:
    if isinstance(obj, list):
        return [from_json(v, project_id, server) for v in obj]
    elif isinstance(obj, dict):
        if (val_type := obj.pop("_type", None)) is None:
            return {k: from_json(v, project_id, server) for k, v in obj.items()}
        elif val_type == "ObjectRecord":
            return ObjectRecord(
                {k: from_json(v, project_id, server) for k, v in obj.items()}
            )
        elif val_type == "CustomWeaveType":
            files = _load_custom_obj_files(project_id, server, obj["files"])
            return custom_objs.decode_custom_obj(
                obj["weave_type"], files, obj.get("load_op")
            )
        else:
            return ObjectRecord(
                {k: from_json(v, project_id, server) for k, v in obj.items()}
            )
    elif isinstance(obj, str) and obj.startswith("weave://"):
        return parse_uri(obj)

    return obj



import os

import pytest

from weave.integrations.integration_utilities import (
    flatten_calls,
    flattened_calls_to_names,
)
from weave.trace.weave_client import WeaveClient
from weave.trace_server.trace_server_interface import CallsFilter


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_dspy_language_models(client: WeaveClient) -> None:
    import dspy

    os.environ["DSP_CACHEBOOL"] = "False"

    gpt3_turbo = dspy.OpenAI(
        model="gpt-3.5-turbo-1106",
        max_tokens=300,
        api_key=os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY"),
    )
    dspy.configure(lm=gpt3_turbo)
    prediction = gpt3_turbo("hello! this is a raw prompt to GPT-3.5")
    expected_prediction = "Hello! How can I assist you today?"
    assert prediction == [expected_prediction]
    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 4

    assert flattened_calls_to_names(flattened_calls) == [
        ("dspy.OpenAI", 0),
        ("dspy.OpenAI.request", 1),
        ("dspy.OpenAI.basic_request", 2),
        ("openai.chat.completions.create", 3),
    ]

    call_1, _ = flattened_calls[0]
    assert call_1.exception is None and call_1.ended_at is not None
    output_1 = call_1.output
    assert output_1[0] == expected_prediction

    call_2, _ = flattened_calls[1]
    assert call_2.exception is None and call_2.ended_at is not None
    output_2 = call_2.output
    assert output_2["choices"][0]["finish_reason"] == "stop"
    assert output_2["choices"][0]["message"]["content"] == expected_prediction
    assert output_2["choices"][0]["message"]["role"] == "assistant"
    assert output_2["model"] == "gpt-3.5-turbo-1106"
    assert output_2["usage"]["completion_tokens"] == 9
    assert output_2["usage"]["prompt_tokens"] == 21
    assert output_2["usage"]["total_tokens"] == 30

    call_3, _ = flattened_calls[2]
    assert call_3.exception is None and call_3.ended_at is not None
    output_3 = call_3.output
    assert output_3["choices"][0]["finish_reason"] == "stop"
    assert output_3["choices"][0]["message"]["content"] == expected_prediction
    assert output_3["choices"][0]["message"]["role"] == "assistant"
    assert output_3["model"] == "gpt-3.5-turbo-1106"
    assert output_3["usage"]["completion_tokens"] == 9
    assert output_3["usage"]["prompt_tokens"] == 21
    assert output_3["usage"]["total_tokens"] == 30
    assert output_2["id"] == output_3["id"]
    assert output_2["created"] == output_3["created"]


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_dspy_inline_signatures(client: WeaveClient) -> None:
    import dspy

    os.environ["DSP_CACHEBOOL"] = "False"

    turbo = dspy.OpenAI(
        model="gpt-3.5-turbo", api_key=os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY")
    )
    dspy.settings.configure(lm=turbo)
    classify = dspy.Predict("sentence -> sentiment")
    prediction = classify(
        sentence="it's a charming and often affecting journey."
    ).sentiment
    expected_prediction = "Positive"
    assert prediction == expected_prediction
    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 6

    assert flattened_calls_to_names(flattened_calls) == [
        ("dspy.Predict", 0),
        ("dspy.Predict.forward", 1),
        ("dspy.OpenAI", 2),
        ("dspy.OpenAI.request", 3),
        ("dspy.OpenAI.basic_request", 4),
        ("openai.chat.completions.create", 5),
    ]

    call_1, _ = flattened_calls[0]
    assert call_1.exception is None and call_1.ended_at is not None
    output_1 = call_1.output
    assert (
        output_1
        == """Prediction(
    sentiment='Positive'
)"""
    )

    call_2, _ = flattened_calls[1]
    assert call_2.exception is None and call_2.ended_at is not None
    output_2 = call_2.output
    assert (
        output_2
        == """Prediction(
    sentiment='Positive'
)"""
    )

    call_3, _ = flattened_calls[2]
    assert call_3.exception is None and call_3.ended_at is not None
    output_3 = call_3.output
    assert output_3[0] == expected_prediction

    call_4, _ = flattened_calls[3]
    assert call_4.exception is None and call_4.ended_at is not None
    output_4 = call_4.output
    assert output_4["choices"][0]["finish_reason"] == "stop"
    assert output_4["choices"][0]["message"]["content"] == expected_prediction
    assert output_4["choices"][0]["message"]["role"] == "assistant"
    assert output_4["model"] == "gpt-3.5-turbo-0125"
    assert output_4["usage"]["completion_tokens"] == 1
    assert output_4["usage"]["prompt_tokens"] == 53
    assert output_4["usage"]["total_tokens"] == 54

    call_5, _ = flattened_calls[4]
    assert call_5.exception is None and call_5.ended_at is not None
    output_5 = call_5.output
    assert output_5["choices"][0]["finish_reason"] == "stop"
    assert output_5["choices"][0]["message"]["content"] == expected_prediction
    assert output_5["choices"][0]["message"]["role"] == "assistant"
    assert output_5["model"] == "gpt-3.5-turbo-0125"
    assert output_5["usage"]["completion_tokens"] == 1
    assert output_5["usage"]["prompt_tokens"] == 53
    assert output_5["usage"]["total_tokens"] == 54
    assert output_5["id"] == output_4["id"]
    assert output_5["created"] == output_4["created"]



import dataclasses
import pickle
import typing

import huggingface_hub
import torch
import transformers

import weave
from weave.legacy.weave import op_def_type
from weave.legacy.weave.ecosystem import pytorch


# This tells Weave how to serialize BaseModelOutput
class HFInternalBaseModelOutputType(weave.types.Type):
    instance_classes = transformers.modeling_outputs.BaseModelOutput

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pickle", binary=True) as f:
            pickle.dump(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pickle", binary=True) as f:
            return pickle.load(f)


class HFInternalPipelineType(weave.types.Type):
    instance_classes = transformers.pipelines.base.Pipeline

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.pickle", binary=True) as f:
            pickle.dump(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.pickle", binary=True) as f:
            return pickle.load(f)


class HFModelType(weave.types.ObjectType):
    def property_types(self):
        return {
            "_id": weave.types.String(),
            "_sha": weave.types.String(),
            "_pipeline_tag": weave.types.optional(weave.types.String()),
            "_tags": weave.types.List(weave.types.String()),
            "_downloads": weave.types.Int(),
            "_likes": weave.types.Int(),
            "_library_name": weave.types.optional(weave.types.String()),
            "call": op_def_type.OpDefType(),
        }


@weave.weave_class(weave_type=HFModelType)
@dataclasses.dataclass
class HFModel:
    _id: str
    _sha: str
    _pipeline_tag: str
    _tags: list[str]  # TODO: we need a Tag type
    _downloads: int
    _likes: int
    _library_name: typing.Optional[str]
    # TODO: version?

    # This is the type of of the call op that will be overridden by child-classes.
    # Just putting it here to get nice autocomplete, but there's probably a better
    # and more general way to do this.
    # TODO: Fix
    # call: weave.OpDef

    def tokenizer(self):
        return transformers.AutoTokenizer.from_pretrained(self._id)

    @weave.op()
    def readme(self) -> weave.legacy.weave.ops.Markdown:
        readme = huggingface_hub.hf_hub_download(self._id, "README.md")
        # quick hack: remove the metadata header from the readme
        readme_contents = ""
        break_count = 0
        for l in open(readme).readlines():
            if l.startswith("---"):
                break_count += 1
            elif break_count > 1:
                readme_contents += l
        # if we failed to parse out the header
        if len(readme_contents) < 1:
            readme_contents = open(readme).read()
        return weave.legacy.weave.ops.Markdown(readme_contents)

    @weave.op()
    def id(self) -> str:
        return self._id

    @weave.op()
    def sha(self) -> str:
        return self._sha

    @weave.op()
    def pipeline_tag(self) -> str:
        return self._pipeline_tag

    @weave.op()
    def tags(self) -> list[str]:
        return self._tags

    @weave.op()
    def downloads(self) -> int:
        return self._downloads

    @weave.op()
    def likes(self) -> int:
        return self._likes

    @weave.op()
    def library_name(self) -> typing.Optional[str]:
        return self._library_name


HFModelType.instance_classes = HFModel


class BaseModelOutputType(weave.types.ObjectType):
    def property_types(self):
        return {
            "_model": HFModelType(),
            "model_input": weave.types.String(),
            "_encoded_model_input": pytorch.TorchTensorType(),
            "model_output": HFInternalBaseModelOutputType(),
        }


class ModelOutputAttentionType(weave.types.ObjectType):
    def property_types(self):
        return {
            "model_output": BaseModelOutputType(),
            "_attention": weave.types.List(pytorch.TorchTensorType()),
        }


@weave.weave_class(weave_type=BaseModelOutputType)
@dataclasses.dataclass
class BaseModelOutput:
    _model: HFModel
    model_input: str
    _encoded_model_input: torch.Tensor
    model_output: transformers.modeling_outputs.BaseModelOutput


@weave.weave_class(weave_type=ModelOutputAttentionType)
@dataclasses.dataclass
class ModelOutputAttention:
    model_output: BaseModelOutput
    _attention: list[torch.Tensor]


class FullPipelineOutputType(weave.types.ObjectType):
    def property_types(self):
        return {
            "_model": HFModelType(),
            "model_input": weave.types.String(),
            "model_output": weave.types.List(weave.types.TypedDict({})),
        }


@weave.weave_class(weave_type=FullPipelineOutputType)
@dataclasses.dataclass
class FullPipelineOutput:
    _model: HFModel
    model_input: str
    model_output: typing.Any

    @weave.op(output_type=ModelOutputAttentionType())
    def attention(self):
        pipeline = weave.use(self._model.pipeline())
        tokenizer = pipeline.tokenizer
        # re-initialize model with output_attentions=True
        model = pipeline.model.__class__.from_pretrained(
            self._model._id, output_attentions=True
        )
        encoded_input = tokenizer.encode(self.model_input, return_tensors="pt")
        model_output = model(encoded_input)
        bmo = BaseModelOutput(
            self._model, self.model_input, encoded_input, model_output
        )
        return ModelOutputAttention(bmo, model_output[-1])



# Put this code in the top of assign_type in weave_types.py to log
# a bunch of types. Then use this test to test profile assignment.
#
# print("FROM TO", next_type, self)
# try:
#     open("/tmp/assigns.jsonl", "a").write(
#         "%s\n" % json.dumps({"from": next_type.to_dict(), "to": self.to_dict()})
#     )
# except TypeError:
#     pass


import cProfile
import json
import time

import pytest

import weave


def do_assigns(assigns):
    for assign_from, assign_to in assigns:
        assign_to.assign_type(assign_from)


@pytest.mark.skip("This is a performance test")
def test_assign_perf():
    assigns = []
    with open("./assigns.jsonl") as f:
        for line in f:
            parsed = json.loads(line)
            assigns.append(
                (
                    weave.types.TypeRegistry.type_from_dict(parsed["from"]),
                    weave.types.TypeRegistry.type_from_dict(parsed["to"]),
                )
            )
    start_time = time.time()
    with cProfile.Profile() as pr:
        for i in range(10):
            do_assigns(assigns)
    pr.dump_stats("assigns.prof")
    print("TIME", time.time() - start_time)
    print("LEN", len(assigns))
    assert 1 == 2
